import sys
import builtins
from importlib import import_module

builtins._ASTROPY_SETUP_ = True

# DO NOT CHANGE ANYTHING (including default units) !!!

if 'astropy.units' in sys.modules:
    u = sys.modules['astropy.units']
else:
    u = import_module('astropy.units')

# DO NOT CHANGE THIS!!!
MASS_UNIT = u.kg
TEMPERATURE_UNIT = u.K
DISTANCE_UNIT = u.m
TIME_UNIT = u.s
ARC_UNIT = u.rad
PERIOD_UNIT = u.d
VELOCITY_UNIT = DISTANCE_UNIT / TIME_UNIT
ACCELERATION_UNIT = DISTANCE_UNIT / TIME_UNIT ** 2
LOG_ACCELERATION_UNIT = u.dex(ACCELERATION_UNIT)
FREQUENCY_UNIT = u.Hz
ANGULAR_FREQUENCY_UNIT = u.rad / u.s
LUMINOSITY_UNIT = u.W
RADIANCE_UNIT = u.W / (u.m ** 2 * u.sr)

# astropy units to avoid annoying undefined warning across the code
deg = u.deg
degree = u.degree
rad = u.rad
km = u.km
solMass = u.solMass
deg_C = u.deg_C
m = u.m
cm = u.cm
d = u.d
s = u.s
W = u.W
solRad = u.solRad
K = u.K
dimensionless_unscaled = u.dimensionless_unscaled
kg = u.kg
dex = u.dex
mag = u.mag
mmag = u.mmag
pc = u.pc
au = u.au

Unit = u.Unit
Quantity = u.quantity.Quantity
Dex = u.Dex

# DEFAULT PARAMETER UNITS -- DO NOT CHANGE!!! --------------------------------------------------------------------------


class BaseUnits(object):
    @classmethod
    def __iter__(cls):
        for varname in cls.__dict__.keys():
            if str(varname).startswith("_"):
                continue
            _unit = getattr(cls, varname)
            if isinstance(_unit, Unit):
                _unit = _unit.to_string()
            yield varname, _unit

    def __getitem__(self, item):
        return getattr(self, item)

    def as_dict(self):
        _dict_repr = dict()
        for key, val in self:
            if isinstance(val, Unit):
                _val_repr = val.to_string()
            elif isinstance(val, BaseUnits):
                _val_repr = val.as_dict()
            elif isinstance(val, bool):
                _val_repr = 'boolean'
            else:
                _val_repr = str(val)

            _dict_repr[key] = _val_repr
        return _dict_repr


# DEFAULT ELISa INNER UNITS (ALL USER INPUTS ARE CONVERTED TO THIS) ----------------------------------------------------

DEFAULT_INCLINATION_UNIT = ARC_UNIT
DEFAULT_PERIOD_UNIT = PERIOD_UNIT
DEFAULT_GAMMA_UNIT = VELOCITY_UNIT


class _DefaultSpotUnits(BaseUnits):
    longitude = ARC_UNIT
    latitude = ARC_UNIT
    angular_radius = ARC_UNIT
    temperature_factor = dimensionless_unscaled
    discretization_factor = ARC_UNIT


class _DefaultPulsationsUnits(BaseUnits):
    l = dimensionless_unscaled
    m = dimensionless_unscaled
    amplitude = VELOCITY_UNIT
    frequency = FREQUENCY_UNIT
    start_phase = ARC_UNIT
    mode_axis_theta = ARC_UNIT
    mode_axis_phi = ARC_UNIT
    temperature_perturbation_phase_shift = ARC_UNIT
    horizontal_to_radial_amplitude_ratio = dimensionless_unscaled
    temperature_amplitude_factor = dimensionless_unscaled
    tidally_locked = bool


class _DefaultStarUnits(BaseUnits):
    mass = MASS_UNIT
    t_eff = TEMPERATURE_UNIT
    metallicity = dimensionless_unscaled
    gravity_darkening = dimensionless_unscaled
    albedo = dimensionless_unscaled
    discretization_factor = ARC_UNIT
    polar_log_g = LOG_ACCELERATION_UNIT
    equivalent_radius = DISTANCE_UNIT
    limb_darkening_coefficients = dimensionless_unscaled
    spots = _DefaultSpotUnits()
    pulsations = _DefaultPulsationsUnits()


class _DefaultSystemUnits(BaseUnits):
    inclination = DEFAULT_INCLINATION_UNIT
    period = DEFAULT_PERIOD_UNIT
    gamma = DEFAULT_GAMMA_UNIT
    phase_shift = dimensionless_unscaled
    additional_light = dimensionless_unscaled
    t0 = DEFAULT_PERIOD_UNIT
    distance = DISTANCE_UNIT


class _DefaultBinarySystemUnits(BaseUnits):
    class __DefaultBinarySystemUnits(BaseUnits):
        inclination = _DefaultSystemUnits.inclination
        period = _DefaultSystemUnits.period
        eccentricity = dimensionless_unscaled
        argument_of_periastron = ARC_UNIT
        gamma = _DefaultSystemUnits.gamma
        phase_shift = _DefaultSystemUnits.phase_shift
        additional_light = _DefaultSystemUnits.additional_light
        primary_minimum_time = _DefaultSystemUnits.t0
        semi_major_axis = DISTANCE_UNIT
        distance = _DefaultSystemUnits.distance
        mass_ratio = dimensionless_unscaled

    class __DefaultBinaryComponentUnits(BaseUnits):
        mass = _DefaultStarUnits.mass
        t_eff = _DefaultStarUnits.t_eff
        surface_potential = dimensionless_unscaled
        synchronicity = dimensionless_unscaled
        metallicity = _DefaultStarUnits.metallicity
        gravity_darkening = _DefaultStarUnits.gravity_darkening
        albedo = _DefaultStarUnits.albedo
        discretization_factor = _DefaultStarUnits.discretization_factor
        limb_darkening_coefficients = _DefaultStarUnits.limb_darkening_coefficients
        spots = _DefaultSpotUnits()
        pulsations = _DefaultPulsationsUnits()

    system = __DefaultBinarySystemUnits()
    component = __DefaultBinaryComponentUnits()
    primary = __DefaultBinaryComponentUnits()
    secondary = __DefaultBinaryComponentUnits()


class _DefaultSingleSystemUnits(BaseUnits):
    class __DefaultSingleSystemUnits(BaseUnits):
        inclination = _DefaultSystemUnits.inclination
        rotation_period = _DefaultSystemUnits.period
        reference_time = _DefaultSystemUnits.t0
        phase_shift = _DefaultSystemUnits.phase_shift
        additional_light = _DefaultSystemUnits.additional_light
        gamma = _DefaultSystemUnits.gamma
        distance = _DefaultSystemUnits.distance

    class __DefaultSingleSystemStarUnits(BaseUnits):
        mass = _DefaultStarUnits.mass
        t_eff = _DefaultStarUnits.t_eff
        polar_log_g = _DefaultStarUnits.polar_log_g
        metallicity = _DefaultStarUnits.metallicity
        gravity_darkening = _DefaultStarUnits.gravity_darkening
        discretization_factor = _DefaultStarUnits.discretization_factor
        equivalent_radius = _DefaultStarUnits.equivalent_radius
        limb_darkening_coefficients = _DefaultStarUnits.limb_darkening_coefficients
        spots = _DefaultSpotUnits()
        pulsations = _DefaultPulsationsUnits()

    system = __DefaultSingleSystemUnits()
    star = __DefaultSingleSystemStarUnits()


DefaultSpotUnits = _DefaultSpotUnits()
DefaultPulsationsUnits = _DefaultPulsationsUnits()
DefaultStarUnits = _DefaultStarUnits()
DefaultBinarySystemUnits = _DefaultBinarySystemUnits()
DefaultSingleSystemUnits = _DefaultSingleSystemUnits()

default_unit_map = {
    'SingleSystem': DefaultSingleSystemUnits.system,
    'BinarySystem': DefaultBinarySystemUnits.system,
    'Star': DefaultStarUnits,
    'Spot': DefaultSpotUnits,

}

# DEFAULT ELISa OUTTER/USER INPUT UNITS (MORE CONVENIENT FOR USER BUT NOT SO MUCH FOR PROGRAMMER) ----------------------

DEFAULT_INCLINATION_INPUT_UNIT = deg
DEFAULT_PERIOD_INPUT_UNIT = d
DEFAULT_GAMMA_INPUT_UNIT = m/s
DISTANCE_TO_OBS_INPUT_UNIT = pc


class _DefaultSpotInputUnits(BaseUnits):
    longitude = deg
    latitude = deg
    angular_radius = deg
    temperature_factor = dimensionless_unscaled
    discretization_factor = deg


class _DefaultPulsationsInputUnits(BaseUnits):
    l = dimensionless_unscaled
    m = dimensionless_unscaled
    amplitude = VELOCITY_UNIT
    frequency = u.d ** (-1)
    start_phase = ARC_UNIT
    mode_axis_theta = deg
    mode_axis_phi = deg
    temperature_perturbation_phase_shift = ARC_UNIT
    horizontal_to_radial_amplitude_ratio = dimensionless_unscaled
    temperature_amplitude_factor = dimensionless_unscaled
    tidally_locked = bool


class _DefaultStarInputUnits(BaseUnits):
    mass = solMass
    t_eff = K
    metallicity = dimensionless_unscaled
    gravity_darkening = dimensionless_unscaled
    albedo = dimensionless_unscaled
    discretization_factor = deg
    polar_log_g = u.dex(cm / s ** 2)
    equivalent_radius = solRad
    limb_darkening_coefficients = dimensionless_unscaled
    spots = _DefaultSpotInputUnits()
    pulsations = _DefaultPulsationsInputUnits()


class _DefaultSystemInputUnits(BaseUnits):
    inclination = DEFAULT_INCLINATION_INPUT_UNIT
    period = DEFAULT_PERIOD_INPUT_UNIT
    gamma = DEFAULT_GAMMA_INPUT_UNIT
    phase_shift = dimensionless_unscaled
    additional_light = dimensionless_unscaled
    t0 = DEFAULT_PERIOD_INPUT_UNIT
    distance = DISTANCE_TO_OBS_INPUT_UNIT


class _DefaultBinarySystemInputUnits(BaseUnits):
    class __DefaultBinaryComponentInputUnits(BaseUnits):
        mass = _DefaultStarInputUnits.mass
        t_eff = _DefaultStarInputUnits.t_eff
        surface_potential = dimensionless_unscaled
        synchronicity = dimensionless_unscaled
        metallicity = _DefaultStarInputUnits.metallicity
        gravity_darkening = _DefaultStarInputUnits.gravity_darkening
        albedo = _DefaultStarInputUnits.albedo
        discretization_factor = _DefaultStarInputUnits.discretization_factor
        limb_darkening_coefficients = _DefaultStarInputUnits.limb_darkening_coefficients
        spots = _DefaultSpotInputUnits()
        pulsations = _DefaultPulsationsInputUnits()

    class __DefaultBinarySystemInputUnits(BaseUnits):
        inclination = _DefaultSystemInputUnits.inclination
        period = _DefaultSystemInputUnits.period
        eccentricity = dimensionless_unscaled
        argument_of_periastron = deg
        gamma = _DefaultSystemInputUnits.gamma
        phase_shift = _DefaultSystemInputUnits.phase_shift
        additional_light = _DefaultSystemInputUnits.additional_light
        primary_minimum_time = _DefaultSystemInputUnits.t0
        semi_major_axis = solRad
        distance = _DefaultSystemInputUnits.distance
        mass_ratio = dimensionless_unscaled

    system = __DefaultBinarySystemInputUnits()
    component = __DefaultBinaryComponentInputUnits()
    primary = __DefaultBinaryComponentInputUnits()
    secondary = __DefaultBinaryComponentInputUnits()


class _DefaultSingleSystemInputUnits(BaseUnits):
    class __DefaultSingleSystemInputUnits(BaseUnits):
        inclination = _DefaultSystemInputUnits.inclination
        rotation_period = _DefaultSystemInputUnits.period
        reference_time = _DefaultSystemInputUnits.t0
        phase_shift = _DefaultSystemInputUnits.phase_shift
        additional_light = _DefaultSystemInputUnits.additional_light
        gamma = _DefaultSystemInputUnits.gamma
        distance = _DefaultSystemInputUnits.distance

    class __DefaultSingleSystemStarInputUnits(BaseUnits):
        mass = _DefaultStarInputUnits.mass
        t_eff = _DefaultStarInputUnits.t_eff
        polar_log_g = _DefaultStarInputUnits.polar_log_g
        metallicity = _DefaultStarInputUnits.metallicity
        gravity_darkening = _DefaultStarInputUnits.gravity_darkening
        discretization_factor = _DefaultStarInputUnits.discretization_factor
        equivalent_radius = _DefaultStarInputUnits.equivalent_radius
        limb_darkening_coefficients = _DefaultStarInputUnits.limb_darkening_coefficients
        spots = _DefaultSpotInputUnits()
        pulsations = _DefaultPulsationsInputUnits()


    system = __DefaultSingleSystemInputUnits()
    star = __DefaultSingleSystemStarInputUnits()


DefaultSpotInputUnits = _DefaultSpotInputUnits()
DefaultPulsationsInputUnits = _DefaultPulsationsInputUnits()
DefaultStarInputUnits = _DefaultStarInputUnits()
DefaultBinarySystemInputUnits = _DefaultBinarySystemInputUnits()
DefaultSingleSystemInputUnits = _DefaultSingleSystemInputUnits()
DefaultSystemInputUnits = _DefaultSystemInputUnits()
DefaultSystemUnits = _DefaultSystemUnits()



import numpy as np

from copy import copy
from scipy.spatial.qhull import Delaunay
from .. import utils as bsutils
from .. orbit import orbit
from ... base import spot
from ... base.types import INT
from ... utils import is_empty
from ... logger import getLogger
from ... import (
    const,
    umpy as up,
    units as u
)
from ... base.surface.faces import (
    initialize_model_container,
    split_spots_and_component_faces,
    set_all_surface_centres,
    calculate_normals,
    mirror_triangulation
)

logger = getLogger("binary_system.surface.faces")


def visibility_test(centres, xlim, component):
    """
    Tests if given faces are visible from the other star.

    :param component: str;
    :param centres: numpy.array;
    :param xlim: visibility threshold in x axis for given component
    :return: numpy.array[bool];
    """
    return centres[:, 0] >= xlim if component == 'primary' else centres[:, 0] <= xlim


def get_visibility_tests(centres, q_test, xlim, component, morphology):
    """
    Method calculates tests for visibilities of faces from other component.
    Used in reflection effect.

    :param centres: np.array; of face centres
    :param q_test: use_quarter_star_test
    :param xlim: visibility threshold in x axis for given component
    :param component: `primary` or `secondary`
    :param morphology: str;
    :return: visual tests for normal and symmetrical star
    """
    if q_test:
        y_test, z_test = centres[:, 1] > 0, centres[:, 2] > 0
        # this branch is activated in case of clean surface where symmetries can be used
        # excluding quadrants that can be mirrored using symmetries
        quadrant_exclusion = up.logical_or(y_test, z_test) if morphology == 'over-contfact' \
            else np.array([True] * len(centres))

        single_quadrant = up.logical_and(y_test, z_test)
        # excluding faces on far sides of components
        test1 = visibility_test(centres, xlim, component)
        # this variable contains faces that can seen from base symmetry part of the other star
        vis_test = up.logical_and(test1, quadrant_exclusion)
        vis_test_symmetry = up.logical_and(test1, single_quadrant)

    else:
        vis_test = centres[:, 0] >= xlim if component == 'primary' else centres[:, 0] <= xlim
        vis_test_symmetry = None

    return vis_test, vis_test_symmetry


def faces_visibility_x_limits(primary_polar_radius, secondary_polar_radius, components_distance):
    """
    Returns x coordinates of `primary` and `secondary` surface elements which can be visible from the other star.

    :param primary_polar_radius: float;
    :param secondary_polar_radius: float;
    :param components_distance: float; in SMA
    :return: Tuple; x_min for primary, x_max for secondary
    """
    # this section calculates the visibility of each surface face
    # don't forget to treat system visibility of faces on the same star in over-contact system

    # if stars are too close and with too different radii, you can see more (less) than a half of the stellar
    # surface, calculating excess angle

    primary_polar_r, secondary_polar_r = primary_polar_radius, secondary_polar_radius
    sin_theta = up.abs(primary_polar_r - secondary_polar_r) / components_distance
    x_corr_primary, x_corr_secondary = primary_polar_r * sin_theta, secondary_polar_r * sin_theta

    # visibility of faces is given by their x position
    xlim = {}
    (xlim['primary'], xlim['secondary']) = (x_corr_primary, components_distance + x_corr_secondary) \
        if primary_polar_r > secondary_polar_r else (-x_corr_primary, components_distance - x_corr_secondary)
    return xlim


def get_surface_builder_fn(morphology):
    """
    Returns suitable triangulation function depending on morphology.

    :return: callable; method that performs generation surface faces
    """
    return over_contact_system_surface if morphology == "over-contact" else detached_system_surface


def build_faces(system, components_distance, component="all"):
    """
    Function creates faces of the star surface for given components. Faces are evaluated upon points that
    have to be in this time already calculated.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;; instance
    :type components_distance: float;
    :param component: `primary` or `secondary` if not supplied both component are calculated
    :return: system; elisa.binary_system.contaier.OrbitalPositionContainer; instance
    """
    if is_empty(component):
        logger.debug("no component set to build faces")
        return system

    if is_empty(components_distance):
        raise ValueError('Value of `components_distance` was not provided.')

    components = bsutils.component_to_list(component)
    for component in components:
        star = getattr(system, component)
        if star.has_spots():
            build_surface_with_spots(system, components_distance, component)
        else:
            build_surface_with_no_spots(system, components_distance, component)
    return system


def build_surface_with_no_spots(system, components_distance, component="all"):
    """
    Function for building binary star component surfaces without spots.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param components_distance: float;
    :param component: str; `primary` or `secondary` if not supplied both component are calculated
    :return: system; elisa.binary_system.contaier.OrbitalPositionContainer; instance
    """
    components = bsutils.component_to_list(component)

    for component in components:
        star = getattr(system, component)
        # triangulating only one quarter of the star

        triangulated_pts = star.symmetry_points()
        if system.morphology != 'over-contact':
            triangles = detached_system_surface(system, components_distance, triangulated_pts, component)
        else:
            neck = np.max(triangulated_pts[:, 0]) if component == 'primary' else np.min(triangulated_pts[:, 0])
            triangulated_pts = \
                np.append(triangulated_pts, np.array([[neck, 0, 0]]), axis=0)
            triangles = over_contact_system_surface(system, triangulated_pts, component)
            # filtering out triangles containing last point in `points_to_triangulate`
            triangles = triangles[np.array(triangles < star.base_symmetry_points_number).all(1)]

        # filtering out faces on xy an xz planes
        y0_test = np.bitwise_not(np.isclose(triangulated_pts[triangles][:, :, 1], 0).all(1))
        z0_test = np.bitwise_not(np.isclose(triangulated_pts[triangles][:, :, 2], 0).all(1))
        triangles = triangles[up.logical_and(y0_test, z0_test)]

        setattr(star, "base_symmetry_faces_number", INT(np.shape(triangles)[0]))
        # lets exploit axial symmetry and fill the rest of the surface of the star
        star.base_symmetry_faces = triangles
        star.faces = mirror_triangulation(triangles, star.inverse_point_symmetry_matrix)

        base_face_symmetry_vector = up.arange(star.base_symmetry_faces_number)
        star.face_symmetry_vector = up.concatenate([base_face_symmetry_vector for _ in range(4)])
    return system


def build_surface_with_spots(system, components_distance, component="all"):
    """
    Function capable of triangulation of spotty stellar surfaces.
    It merges all surface points, triangulates them and then sorts the resulting surface faces under star or spot.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param components_distance: float;
    :param component: str `primary` or `secondary`
    :return: system; elisa.binary_system.contaier.OrbitalPositionContainer; instance
    """
    components = bsutils.component_to_list(component)
    component_com = {'primary': 0.0, 'secondary': components_distance}
    for component in components:
        star_container = getattr(system, component)
        points, vertices_map = star_container.get_flatten_points_map()

        surface_fn = get_surface_builder_fn(system.morphology)
        surface_fn_kwargs = dict(component=component, points=points, components_distance=components_distance)
        faces = surface_fn(system, **surface_fn_kwargs)
        model, spot_candidates = initialize_model_container(vertices_map)
        model = split_spots_and_component_faces(star_container, points, faces, model,
                                                spot_candidates, vertices_map, component_com[component])
        spot.remove_overlaped_spots_by_vertex_map(star_container, vertices_map)
        spot.remap_surface_elements(star_container, model, points)
    return system


def detached_system_surface(system, components_distance, points=None, component="all"):
    """
    Calculates surface faces from the given component's points in case of detached or semi-contact system.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param components_distance: float;
    :param points: numpy.array;
    :param component: str;
    :return: numpy.array; N x 3 array of vertices indices
    """
    component_instance = getattr(system, component)
    if points is None:
        points = component_instance.points

    if not np.any(points):
        raise ValueError(f"{component} component, with class instance name {component_instance.name} do not "
                         "contain any valid surface point to triangulate")
    # there is a problem with triangulation of near over-contact system, delaunay is not good with pointy surfaces
    critical_pot = system.primary.critical_surface_potential if component == 'primary' \
        else system.secondary.critical_surface_potential
    potential = system.primary.surface_potential if component == 'primary' \
        else system.secondary.surface_potential
    if potential - critical_pot > 0.01:
        logger.debug(f'triangulating surface of {component} component using standard method')
        triangulation = Delaunay(points)
        triangles_indices = triangulation.convex_hull
    else:
        logger.debug(f'surface of {component} component is near or at critical potential; therefore custom '
                     f'triangulation method for (near)critical potential surfaces will be used')
        # calculating closest point to the barycentre
        r_near = np.max(points[:, 0]) if component == 'primary' else np.min(points[:, 0])
        # projection of component's far side surface into ``sphere`` with radius r1

        points_to_transform = copy(points)
        if component == 'secondary':
            points_to_transform[:, 0] -= components_distance
        projected_points = \
            r_near * points_to_transform / np.linalg.norm(points_to_transform, axis=1)[:, None]
        if component == 'secondary':
            projected_points[:, 0] += components_distance

        triangulation = Delaunay(projected_points)
        triangles_indices = triangulation.convex_hull

    return triangles_indices


def over_contact_system_surface(system, points=None, component="all", **kwargs):
    # do not remove kwargs, keep compatible interface w/ detached where components distance has to be provided
    # in this case, components distance is sinked in kwargs and not used
    """
    Calculates surface faces from the given component's points in case of over-contact system.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param points: numpy.array; - points to triangulate
    :param component: str; `primary` or `secondary`
    :return: numpy.array; N x 3 array of vertice indices
    """
    del kwargs

    component_instance = getattr(system, component)
    if up.isnan(points).any():
        raise ValueError(f"{component} component, with class instance name {component_instance.name} "
                         f"contain any valid point to triangulate")
    # calculating position of the neck
    neck_x = np.max(points[:, 0]) if component == 'primary' else np.min(points[:, 0])

    projected_points = points.copy()
    projected_points[:, 0] -= 1 if component == 'secondary' else 0
    projected_points = neck_x * projected_points / np.linalg.norm(projected_points, axis=1)[:, None]

    triangulation = Delaunay(projected_points)
    triangles_indices = triangulation.convex_hull

    # removal of faces on top of the neck
    neck_test = ~(up.equal(points[triangles_indices][:, :, 0], neck_x).all(-1))
    new_triangles_indices = triangles_indices[neck_test]

    return new_triangles_indices


def compute_all_surface_areas(system, component):
    """
    Compute surface are of all faces (spots included).

    :param system: elisa.binary_system.container.OrbitalPositionContainer; instance
    :param component: str; `primary` or `secondary`
    :return: system; elisa.binary_system.contaier.OrbitalPositionContainer; instance
    """
    if is_empty(component):
        logger.debug("no component set to build surface areas")
        return

    components = bsutils.component_to_list(component)
    for component in components:
        star = getattr(system, component)
        logger.debug(f'computing surface areas of component: '
                     f'{star} / name: {star.name}')
        star.calculate_all_areas()
    return system


def build_faces_orientation(system, components_distance, component="all"):
    """
    Compute face orientation (normals) for each face.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param component: str; `primary` or `secondary`
    :param components_distance: float; orbit with misaligned pulsations, where pulsation axis drifts with star
    :return: system; elisa.binary_system.contaier.OrbitalPositionContainer; instance
    """
    if is_empty(component):
        logger.debug("no component set to build face orientation")
        return system

    component = bsutils.component_to_list(component)
    com_x = {'primary': 0.0, 'secondary': components_distance}

    for _component in component:
        star = getattr(system, _component)
        set_all_surface_centres(star)
        set_all_normals(star, com=com_x[_component])
    return system


def set_all_normals(star_container, com):
    """
    Function calculates normals for each face of given body (including spots) and assign it to object.

    :param star_container: instance of container to set normals on;
    :param com: numpy.array;
    :param star_container: instance of container to set normals on;
    """
    points, faces, cntrs = star_container.points, star_container.faces, star_container.face_centres
    if star_container.symmetry_test():
        normals1 = calculate_normals(star_container.symmetry_points(),
                                     star_container.symmetry_faces(faces),
                                     star_container.symmetry_faces(cntrs), com)
        normals2 = normals1 * np.array([1.0, -1.0,  1.0])
        normals3 = normals1 * np.array([1.0, -1.0, -1.0])
        normals4 = normals1 * np.array([1.0,  1.0, -1.0])
        star_container.normals = np.concatenate((normals1, normals2, normals3, normals4), axis=0)
    else:
        star_container.normals = calculate_normals(points, faces, cntrs, com)

    if star_container.has_spots() and not star_container.is_flat():
        for spot_index in star_container.spots:
            star_container.spots[spot_index].normals = calculate_normals(star_container.spots[spot_index].points,
                                                                         star_container.spots[spot_index].faces,
                                                                         star_container.spots[spot_index].face_centres,
                                                                         com)
    return star_container


def build_velocities(system, components_distance, component='all'):
    """
    Function calculates velocity vector for each face relative to the system's centre of mass.

    :param system: elisa.binary_system.container.SystemContainer;
    :param components_distance: float;
    :param component: str;
    :return: elisa.binary_system.container.SystemContainer;
    """
    if is_empty(component):
        logger.debug("no component set to build face orientation")
        return system

    component = bsutils.component_to_list(component)
    com_x = {'primary': np.array([0.0, 0.0, 0.0]), 'secondary': np.array([components_distance, 0.0, 0.0])}

    velocities = orbit.create_orb_vel_vectors(system, components_distance)

    orb_period = (system.period * u.DefaultBinarySystemUnits.system.period).to(u.TIME_UNIT).value
    omega_orb = np.array([0, 0, const.FULL_ARC / orb_period])

    for _component in component:
        star = getattr(system, _component)
        points = (star.points - com_x[_component][None, :]) * system.semi_major_axis
        omega = star.synchronicity * omega_orb

        # orbital velocity + rotational velocity
        p_velocities = velocities[_component] + np.cross(omega[None, :], points, axisa=1)
        star.velocities = np.mean(p_velocities[star.faces], axis=1)

        if star.has_spots():
            for spot_inst in star.spots.values():
                points = (spot_inst.points - com_x[_component][None, :]) * system.semi_major_axis
                p_velocities = velocities[_component] + np.cross(omega[None, :], points, axisa=1)
                spot_inst.velocities = np.mean(p_velocities[spot_inst.faces], axis=1)

    return system



import os
import sys
import configparser
import os.path as op
import numpy as np

from pathlib import Path


class DefaultSettings(object):
    LOG_CONFIG = op.join(op.dirname(op.abspath(__file__)), 'logging_schemas/default.json')
    SUPPRESS_WARNINGS = False
    SUPPRESS_LOGGER = None
    HOME = op.expanduser(op.join("~", '.elisa'))

    # physics
    REFLECTION_EFFECT = True
    REFLECTION_EFFECT_ITERATIONS = 2
    LIMB_DARKENING_LAW = 'cosine'
    PULSATION_MODEL = 'uniform'
    DEFAULT_TEMPERATURE_PERTURBATION_PHASE_SHIFT = np.pi / 3.0
    SURFACE_DISPLACEMENT_TOL = 1e-2
    RV_METHOD = 'kinematic'
    RV_LAMBDA_INTERVAL = (5500, 5600)

    # computational
    MAX_DISCRETIZATION_FACTOR = 8
    MIN_DISCRETIZATION_FACTOR = 3
    DEFAULT_DISCRETIZATION_FACTOR = 5
    NUMBER_OF_THREADS = 1
    NUMBER_OF_PROCESSES = -1  # int(os.cpu_count())
    NUMBER_OF_MCMC_PROCESSES = -1
    MAX_NU_SEPARATION = 0.08
    MAX_D_FLUX = 2e-4
    MAX_SPOT_D_LONGITUDE = np.pi / 180.0  # in radians
    MIN_POINTS_IN_ECLIPSE = 35
    MAX_SOLVER_ITERS = 100
    MAX_CURVE_DATA_POINTS = 300
    MESH_GENERATOR = 'auto'
    DEFORMATION_TOL = 0.05
    MCMC_SAVE_INTERVAL = 1800
    USE_SINGLE_LD_COEFFICIENTS = False
    USE_INTERPOLATION_APPROXIMATION = True
    USE_SYMMETRICAL_COUNTERPARTS_APPROXIMATION = True
    USE_SIMILAR_NEIGHBOURS_APPROXIMATION = True


    TIMER = 0.0

    # support data
    PASSBAND_TABLES = op.join(op.dirname(op.abspath(__file__)), op.pardir, "passband")
    LD_TABLES = op.join(HOME, "limbdarkening", "ld")
    CK04_ATM_TABLES = op.join(HOME, "atmosphere", "ck04")
    K93_ATM_TABLES = op.join(HOME, "atmosphere", "k93")
    ATM_ATLAS = "ck04"
    ATLAS_TO_BASE_DIR = {
        "castelli": CK04_ATM_TABLES,
        "castelli-kurucz": CK04_ATM_TABLES,
        "ck": CK04_ATM_TABLES,
        "ck04": CK04_ATM_TABLES,
        "kurucz": K93_ATM_TABLES,
        "k": K93_ATM_TABLES,
        "k93": K93_ATM_TABLES
    }
    CUDA = False
    MAGNITUDE_SYSTEM = 'vega'


class SettingsManager(object):
    @staticmethod
    def run():
        def _create_configuration_path_input(_directory):
            if not op.isdir(_directory):
                os.makedirs(_directory, exist_ok=True)

        def _write_ini_file(_config, _path):
            with open(_path, "w") as _f:
                _config.write(_f)

        def _system_exit(exit_code):
            sys.exit(exit_code)

        print("========================== ELISa - first time run ==========================")
        print("It seems that you are using ELISa for the first time.\n"
              "ELISa requires your attention to do a minimal necessary configuration.\n"
              "Please type your configuration preferenceies in following wizzard.\n"
              "If you are power user and you would like to configure it manually, \n"
              "than abort this wizzard and put your configuration in following posible locations: \n"
              "\t- set the environment variable ELISA_CONFIG with absolute path to you config file, or \n "
              "\t- add conf/elisa_conf.ini under your virtualenv root, or \n"
              "\t- add ~/.elisa/config.ini file.\n"
              "For more cofigurable options take a look at: \n"
              "\t\thttps://github.com/mikecokina/elisa/blob/dev/src/elisa/conf/elisa_conf_docs.ini")

        proceed = input("Do you want to continue [y/N]: ")
        if not proceed.lower() == "y":
            _system_exit(0)

        # sections
        section_general = "general"
        section_support = "support"

        # create default ini and add sections
        config = configparser.ConfigParser()
        config.add_section(section_general)
        config.add_section(section_support)

        # configuration file
        default_config_path = op.join(DefaultSettings.HOME, 'config.ini')
        file_prompt = f"Absolute path to configuration file\n" \
                      f"[ hit enter for default {default_config_path} ]: "
        configuration_file = input(file_prompt) or default_config_path
        _create_configuration_path_input(str(Path(configuration_file).parent))
        config.set(section_general, "home", str(Path(configuration_file).parent))

        # ld tables path
        ld_prompt = f"Absolute path to limb darkening tables (will be downloaded)\n" \
                    f"[ default {DefaultSettings.LD_TABLES} ]: "
        ld_tables = input(ld_prompt) or DefaultSettings.LD_TABLES
        _create_configuration_path_input(ld_tables)
        config.set(section_support, "ld_tables", ld_tables)

        # atlas
        atlas_prompt = f"Choose default atmosphere model to use \n" \
                       f"\t type ck04 for catelli-kurucz 2004\n" \
                       f"\t type k93 for kurucz 1993\n" \
                       f"[ default  {DefaultSettings.ATM_ATLAS} ]: "
        atm_atlas = input(atlas_prompt) or DefaultSettings.ATM_ATLAS
        if atm_atlas not in ["ck04", "k93"]:
            raise ValueError("Invalid input for atmosphere atlas. Type option ck04 or k93.")
        config.set(section_support, "atlas", atm_atlas)

        # castelli_kurucz_04_atm_tables
        ck04_prompt = f"Absolute path to store castelli-kurucz 2004 atmosphere models (will be downloaded)\n" \
                      f"[ default {DefaultSettings.CK04_ATM_TABLES} ]: "
        ck04_atm_tables = input(ck04_prompt) or DefaultSettings.CK04_ATM_TABLES
        _create_configuration_path_input(ck04_atm_tables)
        config.set(section_support, "castelli_kurucz_04_atm_tables", ck04_atm_tables)

        # kurucz_93_atm_tables
        k93_prompt = f"Absolute path to store kurucz 1993 atmosphere models (will be downloaded)\n" \
                     f"[ default {DefaultSettings.K93_ATM_TABLES} ]: "
        k93_atm_tables = input(k93_prompt) or DefaultSettings.K93_ATM_TABLES
        _create_configuration_path_input(k93_atm_tables)
        config.set(section_support, "kurucz_93_atm_tables", k93_atm_tables)

        # write file
        _write_ini_file(config, default_config_path)




import numpy as np
import mpl_toolkits.mplot3d.axes3d as axes3d

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import matplotlib.gridspec as gridspec

from matplotlib import cm
from .. import (
    umpy as up,
    utils,
    units
)
from .. import units as u
from . import utils as gutils


CMAPS = {'temperature': cm.jet_r,
         'velocity': cm.jet,
         'radial_velocity': cm.jet,
         'v_r_perturbed': cm.jet,
         'v_horizontal_perturbed': cm.jet,
         'gravity_acceleration': cm.jet,
         'horizontal_acceleration': cm.jet,
         'radius': cm.jet,
         'normal_radiance': cm.hot,
         'radiance': cm.hot,
         'horizontal_displacement': cm.jet,
         }


def orbit(**kwargs):
    """
    Graphics part of the function for quick 2D plot of the orbital motion in the orbital plane.

    :param kwargs: Dict;
    :**kwargs options**:
        * **start_phase** * -- float; starting phase for the plot
        * **stop_phase** * -- float; finishing phase for the plot
        * **number_of_points** * -- int; number of points in the plot
        * **axis_units** * -- Union[astropy.unit, 'str']; specifying axis unit, use astropy units or `dimensionless`
                              or `SMA` (semi-major axis) units for axis scale
        * **frame_of_reference** * -- str; `barycentric` or `primary`
    """
    unit = str(kwargs['axis_units'])
    if kwargs['axis_units'] == units.dimensionless_unscaled:
        x_label, y_label = 'x/[SMA]', 'y/[SMA]'
    else:
        x_label, y_label = r'x/' + unit, r'y/' + unit

    f = plt.figure()
    ax = f.add_subplot(111)
    ax.grid()
    if kwargs['frame_of_reference'] == 'barycentric':
        x1, y1 = kwargs['x1_data'], kwargs['y1_data']
        x2, y2 = kwargs['x2_data'], kwargs['y2_data']
        ax.plot(x1, y1, label='primary')
        ax.plot(x2, y2, label='secondary')
        ax.scatter([0], [0], c='black', s=4)
    elif kwargs['frame_of_reference'] == 'primary':
        x, y = kwargs['x_data'], kwargs['y_data']
        ax.plot(x, y, label='secondary')
        # ax.scatter(x[0], y[0], c='r')
        ax.scatter([0], [0], c='b', label='primary')

    if kwargs['legend']:
        ax.legend(loc=1)
    ax.set_aspect('equal')
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    plt.subplots_adjust(top=0.98, right=0.98)
    return f if kwargs['return_figure_instance'] else plt.show()


def equipotential(**kwargs):
    """
    Plot function for descriptor = `equipotential` in function BinarySystem.plot(). This function plots crossections of
    surface Hill planes in xy, yz or zx plane

    :param kwargs: Dict;
    :**kwargs options**:
        * **plane** * -- str; 'xy' - plane in which surface Hill plane is calculated, planes: 'xy', 'yz', 'zx'
        * **phase** * float; photometric phase in which surface Hill plane is calculated
    """
    x_label, y_label = 'x', 'y'
    if utils.is_plane(kwargs['plane'], 'yz'):
        x_label, y_label = 'y', 'z'
    elif utils.is_plane(kwargs['plane'], 'zx'):
        x_label, y_label = 'x', 'z'

    x_primary, y_primary = kwargs['points_primary'][:, 0], kwargs['points_primary'][:, 1]
    x_secondary, y_secondary = kwargs['points_secondary'][:, 0], kwargs['points_secondary'][:, 1]

    f = plt.figure()
    ax = f.add_subplot(111)

    if kwargs['components_to_plot'] in ['primary', 'both']:
        ax.plot(x_primary, y_primary, label='primary', c=kwargs['colors'][0])
    if kwargs['components_to_plot'] in ['secondary', 'both']:
        ax.plot(x_secondary, y_secondary, label='secondary', c=kwargs['colors'][1])

    lims = ax.get_xlim() - np.mean(ax.get_xlim())
    ax.set_ylim(lims)
    ax.set_aspect('equal', 'box')
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    if kwargs['legend']:
        ax.legend(loc=kwargs['legend_loc'])
    ax.grid()

    return f if kwargs['return_figure_instance'] else plt.show()


def equipotential_single_star(**kwargs):
    """
    Plot function for descriptor = `equipotential` in function SingleSystem.plot(). Calculates zx plane crossection of
    equipotential surface.

    :param kwargs: Dict;
    :**kwargs options**:
        * **axis_unit** * -- astropy.units.solRad - unit in which axis will be displayed, please use
                             astropy.units format, default unit is solar radius
    """

    x, y = kwargs['points'][:, 0], kwargs['points'][:, 1]

    unit = str(kwargs['axis_unit'])
    x_label, y_label = r'x/' + unit, r'z/' + unit

    f = plt.figure()
    ax = f.add_subplot(111)
    ax.grid()
    ax.plot(x, y)
    ax.set_aspect('equal', 'box')
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    return f if kwargs['return_figure_instance'] else plt.show()


def single_star_mesh(**kwargs):
    """
    Plot function for descriptor `mesh`, plots surface mesh of star in SingleStar system

    :param kwargs: Dict;
    :**kwargs options**:
        * **axis_unit** * -- astropy.units.solRad - unit in which axis will be displayed, please use
                            astropy.units format, default unit is solar radius
    """
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.elev = 90 - kwargs['inclination']
    ax.azim = kwargs['azimuth']

    ax.scatter(kwargs['mesh'][:, 0], kwargs['mesh'][:, 1], kwargs['mesh'][:, 2], s=2)
    ax.set_xlim3d(-kwargs['equatorial_radius'], kwargs['equatorial_radius'])
    ax.set_ylim3d(-kwargs['equatorial_radius'], kwargs['equatorial_radius'])
    ax.set_zlim3d(-kwargs['equatorial_radius'], kwargs['equatorial_radius'])
    ax.set_box_aspect([1, 1, 1])
    if kwargs['plot_axis']:
        unit = str(kwargs['axis_unit'])
        x_label, y_label, z_label = r'x/' + unit, r'y/' + unit, r'z/' + unit
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.set_zlabel(z_label)
    else:
        ax.set_axis_off()
    plt.subplots_adjust(left=0.0, right=1.0, top=1.0, bottom=0.0)

    return fig if kwargs['return_figure_instance'] else plt.show()


def binary_mesh(**kwargs):
    """
    Function plots 3D scatter plot of the surface points

    :param kwargs: Dict;
    :**kwargs options**:
        * **phase** * -- float; phase at which to construct plot
        * **components_to_plot** * -- str; component to plot `primary`, `secondary` or `both` (default)
        * **plot_axis** * -- bool; switch the plot axis on/off
        * **inclination** * -- float; elevation of the camera (in degrees)
        * **azimuth** * -- float; azimuth of the camera (in degrees)
    """

    fig = plt.figure(figsize=(7, 7))
    ax = fig.add_subplot(111, projection='3d')
    ax.set_box_aspect([1, 1, 1])
    ax.elev = 90 - kwargs['inclination']
    ax.azim = kwargs['azimuth']
    if kwargs['components_to_plot'] in ['primary', 'both']:
        ax.scatter(kwargs['points_primary'][:, 0], kwargs['points_primary'][:, 1], kwargs['points_primary'][:, 2], s=5,
                   label='primary', alpha=1.0)
    if kwargs['components_to_plot'] in ['secondary', 'both']:
        ax.scatter(kwargs['points_secondary'][:, 0], kwargs['points_secondary'][:, 1], kwargs['points_secondary'][:, 2],
                   s=2, label='secondary', alpha=1.0)
    ax.legend(loc=1)

    x_min, x_max = 0, 0
    if kwargs['components_to_plot'] == 'both':
        x_min = np.min(kwargs['points_primary'][:, 0])
        x_max = np.max(kwargs['points_secondary'][:, 0])
    elif kwargs['components_to_plot'] == 'primary':
        x_min = np.min(kwargs['points_primary'][:, 0])
        x_max = np.max(kwargs['points_primary'][:, 0])
    elif kwargs['components_to_plot'] == 'secondary':
        x_min = np.min(kwargs['points_secondary'][:, 0])
        x_max = np.max(kwargs['points_secondary'][:, 0])

    d = (x_max - x_min) / 2
    ax.set_xlim3d(x_min, x_max)
    ax.set_ylim3d(-d, d)
    ax.set_zlim3d(-d, d)

    if kwargs['plot_axis']:
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
    else:
        ax.set_axis_off()
    plt.subplots_adjust(left=0.0, right=1.0, top=1.0, bottom=0.0)

    return fig if kwargs['return_figure_instance'] else plt.show()


def single_star_surface(**kwargs):
    """
    Plot function for descriptor `surface` in SingleSystem plot function, plots surface of star in SingleStar system

    :param kwargs: Dict;
    :**kwargs options**:
        * **phase** * float -- phase at which plot the system, important for eccentric orbits
        * **components_to_plot** * -- str; `primary`, `secondary` or `both` (default),
        * **normals** * -- bool; plot normals of the surface phases as arrows
        * **edges** * -- bool; highlight edges of surface faces
        * **colormap** * -- str; `gravity_acceleration`, `temperature` or None(default)
        * **plot_axis** * -- bool; if False, axis will be hidden
        * **face_mask_primary** * -- array[bool]; mask to select which faces to display
        * **face_mask_secondary** * -- array[bool]: mask to select which faces to display
        * **elevation** * -- float; in degree - elevation of camera
        * **azimuth** * -- float; camera azimuth
        * **units** * -- str; units of gravity acceleration colormap  `SI` or `cgs`
        * **scale** * -- str; `linear` or `log`
        * **axis_unit** * -- Union[astropy.unit, dimensionless]; - axis units
        * **colorbar_orientation** * -- str; `horizontal` or `vertical` (default)
        * **colorbar** * -- bool; colorabar on/off switchic
        * **colorbar_separation** * -- float; shifting position of the colorbar from its default postition
        * **colorbar_size** * -- float; relative size of the colorbar, default 0.7
    """

    fig = plt.figure(figsize=(7, 7))
    ax = axes3d.Axes3D(fig)
    ax.set_box_aspect([1, 1, 1])
    ax.elev = kwargs['elevation']
    ax.azim = kwargs['azimuth']

    clr = kwargs['surface_color']

    star_plot = ax.plot_trisurf(kwargs['points'][:, 0], kwargs['points'][:, 1], kwargs['points'][:, 2],
                                triangles=kwargs['triangles'], antialiased=True, shade=True, color=clr)
    if kwargs['edges']:
        star_plot.set_edgecolor('black')

    if kwargs['normals']:
        ax.quiver(kwargs['centres'][:, 0], kwargs['centres'][:, 1], kwargs['centres'][:, 2],
                  kwargs['arrows'][:, 0], kwargs['arrows'][:, 1], kwargs['arrows'][:, 2], color='black',
                  length=0.1 * kwargs['equatorial_radius'])

    if kwargs.get('colormap', False):
        cmap = CMAPS[kwargs['colormap']]
        star_plot.set_cmap(cmap=cmap)
        star_plot.set_array(kwargs['cmap'])
        if kwargs['colorbar']:
            colorbar = fig.colorbar(star_plot, shrink=kwargs['colorbar_size'],
                                    orientation=kwargs['colorbar_orientation'],
                                    pad=kwargs['colorbar_separation'])
            set_colorbar_label(colorbar, kwargs['colormap'], kwargs['unit'], kwargs['scale'])

    ax.set_xlim3d(-kwargs['equatorial_radius'], kwargs['equatorial_radius'])
    ax.set_ylim3d(-kwargs['equatorial_radius'], kwargs['equatorial_radius'])
    ax.set_zlim3d(-kwargs['equatorial_radius'], kwargs['equatorial_radius'])

    if kwargs['plot_axis']:
        unit = str(kwargs['axis_unit'])
        x_label, y_label, z_label = r'x/' + unit, r'y/' + unit, r'z/' + unit
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.set_zlabel(z_label)
    else:
        ax.set_axis_off()
    plt.subplots_adjust(left=0.0, right=1.0, top=1.0, bottom=0.0)

    return fig if kwargs['return_figure_instance'] else plt.show()


def binary_surface(**kwargs):
    """
    Function creates plot of binary system components.

    :param kwargs: Dict;
    :**kwargs options**:
        * **phase** * float -- phase at which plot the system, important for eccentric orbits
        * **components_to_plot** * -- str; `primary`, `secondary` or `both` (default),
        * **normals** * -- bool; plot normals of the surface phases as arrows
        * **edges** * -- bool; highlight edges of surface faces
        * **colormap** * -- str; `gravity_acceleration`, `temperature` or None(default)
        * **plot_axis** * -- bool; if False, axis will be hidden
        * **face_mask_primary** * -- array[bool]; mask to select which faces to display
        * **face_mask_secondary** * -- array[bool]: mask to select which faces to display
        * **elevation** * -- float; in degree - elevation of camera
        * **azimuth** * -- float; camera azimuth
        * **unit** * -- str; units of gravity acceleration colormap  `SI` or `cgs`
        * **scale** * -- str; `linear` or `log`
        * **axis_unit** * -- Union[astropy.unit, dimensionless]; - axis units
        * **colorbar_orientation** * -- str; `horizontal` or `vertical` (default)
        * **colorbar** * -- bool; colorabar on/off switchic
        * **colorbar_separation** * -- float; shifting position of the colorbar from its default postition
        * **colorbar_size** * -- float; relative size of the colorbar, default 0.7
    """

    fig = plt.figure(figsize=(7, 7))
    ax = fig.add_subplot(111, projection='3d')
    ax.set_box_aspect([1, 1, 1])
    ax.elev = kwargs['elevation']
    ax.azim = kwargs['azimuth']

    clr = kwargs['surface_colors']

    plot, plot1, plot2 = None, None, None
    if kwargs['components_to_plot'] == 'primary':
        plot = ax.plot_trisurf(
            kwargs['points_primary'][:, 0], kwargs['points_primary'][:, 1],
            kwargs['points_primary'][:, 2], triangles=kwargs['primary_triangles'],
            antialiased=True, shade=True, color=clr[0])

        if kwargs.get('normals', False):
            ax.quiver(
                kwargs['primary_centres'][:, 0], kwargs['primary_centres'][:, 1], kwargs['primary_centres'][:, 2],
                kwargs['primary_arrows'][:, 0], kwargs['primary_arrows'][:, 1], kwargs['primary_arrows'][:, 2],
                color='black', length=0.05)

    elif kwargs['components_to_plot'] == 'secondary':
        plot = ax.plot_trisurf(kwargs['points_secondary'][:, 0], kwargs['points_secondary'][:, 1],
                               kwargs['points_secondary'][:, 2], triangles=kwargs['secondary_triangles'],
                               antialiased=True, shade=True, color=clr[1])

        if kwargs.get('normals', False):
            ax.quiver(kwargs['secondary_centres'][:, 0], kwargs['secondary_centres'][:, 1],
                      kwargs['secondary_centres'][:, 2],
                      kwargs['secondary_arrows'][:, 0], kwargs['secondary_arrows'][:, 1],
                      kwargs['secondary_arrows'][:, 2],
                      color='black', length=0.05)

    elif kwargs['components_to_plot'] == 'both':
        if not kwargs['separate_colormaps']:
            points = up.concatenate((kwargs['points_primary'], kwargs['points_secondary']), axis=0)
            triangles = up.concatenate((kwargs['primary_triangles'],
                                    kwargs['secondary_triangles'] + np.shape(kwargs['points_primary'])[0]), axis=0)

            plot = ax.plot_trisurf(points[:, 0], points[:, 1], points[:, 2], triangles=triangles, antialiased=True,
                                   shade=True, color=clr[0])
        else:
            plot1 = ax.plot_trisurf(kwargs['points_primary'][:, 0], kwargs['points_primary'][:, 1],
                                    kwargs['points_primary'][:, 2], triangles=kwargs['primary_triangles'],
                                    antialiased=True, shade=True, color=clr[0])
            plot2 = ax.plot_trisurf(kwargs['points_secondary'][:, 0], kwargs['points_secondary'][:, 1],
                                    kwargs['points_secondary'][:, 2], triangles=kwargs['secondary_triangles'],
                                    antialiased=True, shade=True, color=clr[1])
        if kwargs.get('normals', False):
            centres = up.concatenate((kwargs['primary_centres'], kwargs['secondary_centres']), axis=0)
            arrows = up.concatenate((kwargs['primary_arrows'], kwargs['secondary_arrows']), axis=0)

            ax.quiver(centres[:, 0], centres[:, 1], centres[:, 2],
                      arrows[:, 0], arrows[:, 1], arrows[:, 2],
                      color='black', length=0.05)

    else:
        raise ValueError('Invalid value of keyword argument `components_to_plot`. '
                         'Expected values are: `primary`, `secondary` or `both`')

    if kwargs.get('edges', False):
        if kwargs['separate_colormaps']:
            plot1.set_edgecolor('black')
            plot2.set_edgecolor('black')
        else:
            plot.set_edgecolor('black')

    if kwargs.get('colormap', False):
        cmap = CMAPS[kwargs['colormap']]
        if kwargs['separate_colormaps']:
            plot1.set_cmap(cmap=cmap)
            plot2.set_cmap(cmap=cmap)
        else:
            plot.set_cmap(cmap=cmap)

        if kwargs['components_to_plot'] == 'primary':
            plot.set_array(kwargs['primary_cmap'])
            if kwargs['colorbar']:
                colorbar = fig.colorbar(plot, shrink=kwargs['colorbar_size'],
                                        orientation=kwargs['colorbar_orientation'],
                                        pad=kwargs['colorbar_separation'])
                set_colorbar_label(colorbar, kwargs['colormap'], kwargs['unit'], kwargs['scale'], extra='primary')
        elif kwargs['components_to_plot'] == 'secondary':
            plot.set_array(kwargs['secondary_cmap'])
            if kwargs['colorbar']:
                colorbar = fig.colorbar(plot, shrink=kwargs['colorbar_size'],
                                        orientation=kwargs['colorbar_orientation'],
                                        pad=kwargs['colorbar_separation'])
                set_colorbar_label(colorbar, kwargs['colormap'], kwargs['unit'], kwargs['scale'], extra='secondary')
        elif kwargs['components_to_plot'] == 'both':
            if not kwargs['separate_colormaps']:
                both_cmaps = up.concatenate((kwargs['primary_cmap'], kwargs['secondary_cmap']), axis=0)
                plot.set_array(both_cmaps)
                if kwargs['colorbar']:
                    colorbar = fig.colorbar(plot, shrink=kwargs['colorbar_size'],
                                            orientation=kwargs['colorbar_orientation'],
                                            pad=kwargs['colorbar_separation'])
                    set_colorbar_label(colorbar, kwargs['colormap'], kwargs['unit'], kwargs['scale'])
            else:
                plot1.set_array(kwargs['primary_cmap'])
                plot2.set_array(kwargs['secondary_cmap'])
                if kwargs['colorbar']:
                    colorbar1 = fig.colorbar(plot1, shrink=kwargs['colorbar_size'],
                                             orientation=kwargs['colorbar_orientation'],
                                             pad=kwargs['colorbar_separation'])
                    set_colorbar_label(
                        colorbar1, kwargs['colormap'], kwargs['unit'], kwargs['scale'], extra='primary'
                    )
                    colorbar2 = fig.colorbar(plot2, shrink=kwargs['colorbar_size'],
                                             orientation=kwargs['colorbar_orientation'],
                                             pad=kwargs['colorbar_separation'])
                    set_colorbar_label(
                        colorbar2, kwargs['colormap'], kwargs['unit'], kwargs['scale'], extra='secondary'
                    )

    if kwargs['plot_axis']:
        unit = str(kwargs['axis_unit'])
        if kwargs['axis_unit'] == units.dimensionless_unscaled:
            x_label, y_label, z_label = 'x', 'y', 'z'
        else:
            x_label, y_label, z_label = r'x/' + unit, r'y/' + unit, r'z/' + unit
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.set_zlabel(z_label)
    else:
        ax.set_axis_off()

    plt.subplots_adjust(left=0.0, right=1.0, top=1.0, bottom=0.0)
    gutils.set_axes_equal(ax)

    return fig if kwargs['return_figure_instance'] else plt.show()


def set_colorbar_label(colorbar, colorbar_name, unit, scale, extra=''):
    lbl = {
        'temperature': 'T',
        'gravity_acceleration': 'g',
        'velocity': 'v',
        'radial_velocity': r'v$_{rad}$',
        'v_r_perturbed': r'$d v_r$',
        'v_horizontal_perturbed': r'$d v_{horizontal}$',
        'normal_radiance': r'I$_{norm}$',
        'radiance': 'I',
        'radius': '$r$',
        'horizontal_displacement': r'$d r_{horizontal}$',
        'horizontal_acceleration': r'$g_{horizontal}$',
    }
    def_unit = {
        'temperature': 'K',
        'gravity_acceleration': r'$m\,s^{-2}$',
        'velocity': r'$m\,s^{-1}$',
        'radial_velocity': r'$m\,s^{-1}$',
        'v_r_perturbed': r'$m\,s^{-1}$',
        'v_horizontal_perturbed': r'$m\,s^{-1}$',
        'normal_radiance': '$W.sr^{-1}.m^{-2}$',
        'radiance': '$W.sr^{-1}.m^{-2}$',
        'radius': '$m$',
        'horizontal_displacement': r'$m$',
        'horizontal_acceleration': r'$m\,s^{-2}$',
    }
    unt = def_unit[colorbar_name] if unit == 'default' else unit
    if scale == 'linear':
        colorbar.set_label(extra + f' {lbl[colorbar_name]}/[{unt}]')
    elif scale == 'log':
        colorbar.set_label(extra + f' log({lbl[colorbar_name]}/[{unt}])')


def single_star_wireframe(**kwargs):
    """
    Plot function for descriptor `wireframe` in SingleSystem, plots wireframe model of single system star

    :param kwargs: Dict;
    :**kwargs options**:
        * **axis_unit** * -- astropy.units.solRad - unit in which axis will be displayed, please use
                                                    astropy.units format, default unit is solar radius
    """
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.elev = 90 - kwargs['inclination']
    ax.azim = kwargs['azimuth']

    star_plot = ax.plot_trisurf(kwargs['mesh'][:, 0], kwargs['mesh'][:, 1], kwargs['mesh'][:, 2],
                                triangles=kwargs['triangles'], antialiased=True, color='none')
    star_plot.set_edgecolor('black')

    ax.set_xlim3d(-kwargs['equatorial_radius'], kwargs['equatorial_radius'])
    ax.set_ylim3d(-kwargs['equatorial_radius'], kwargs['equatorial_radius'])
    ax.set_zlim3d(-kwargs['equatorial_radius'], kwargs['equatorial_radius'])
    ax.set_box_aspect([1, 1, 1])
    if kwargs['plot_axis']:
        unit = str(kwargs['axis_unit'])
        x_label, y_label, z_label = r'x/' + unit, r'y/' + unit, r'z/' + unit
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.set_zlabel(z_label)
    else:
        ax.set_axis_off()
    plt.subplots_adjust(left=0.0, right=1.0, top=1.0, bottom=0.0)

    return fig if kwargs['return_figure_instance'] else plt.show()


def binary_wireframe(**kwargs):
    """
    Function displays wireframe model of the stellar surface

    :param kwargs:
    :**kwargs options**:
        * **phase** * -- float; phase at which to construct plot
        * **components_to_plot** * -- str; component to plot `primary`, `secondary` or `both` (default)
        * **plot_axis** * -- bool; switch the plot axis on/off
        * **inclination** * -- float; elevation of the camera (in degrees)
        * **azimuth** * -- float; azimuth of the camera (in degrees)
    :return:
    """
    fig = plt.figure(figsize=(7, 7))
    ax = fig.add_subplot(111, projection='3d')
    ax.set_box_aspect([1, 1, 1])
    ax.elev = 90 - kwargs['inclination']
    ax.azim = kwargs['azimuth']

    if kwargs['components_to_plot'] == 'primary':
        plot = ax.plot_trisurf(
            kwargs['points_primary'][:, 0], kwargs['points_primary'][:, 1],
            kwargs['points_primary'][:, 2], triangles=kwargs['primary_triangles'],
            antialiased=True, shade=False, color='none')

    elif kwargs['components_to_plot'] == 'secondary':
        plot = ax.plot_trisurf(kwargs['points_secondary'][:, 0], kwargs['points_secondary'][:, 1],
                               kwargs['points_secondary'][:, 2], triangles=kwargs['secondary_triangles'],
                               antialiased=True, shade=False, color='none')

    elif kwargs['components_to_plot'] == 'both':
        points = up.concatenate((kwargs['points_primary'], kwargs['points_secondary']), axis=0)
        triangles = up.concatenate((kwargs['primary_triangles'],
                                    kwargs['secondary_triangles'] + np.shape(kwargs['points_primary'])[0]), axis=0)

        plot = ax.plot_trisurf(points[:, 0], points[:, 1], points[:, 2], triangles=triangles, antialiased=True,
                               shade=False, color='none')

    else:
        raise ValueError('Invalid value of keyword argument `components_to_plot`. '
                         'Expected values are: `primary`, `secondary` or `both`')

    plot.set_edgecolor('black')

    x_min, x_max = 0, 0
    if kwargs['components_to_plot'] == 'both':
        x_min = np.min(kwargs['points_primary'][:, 0])
        x_max = np.max(kwargs['points_secondary'][:, 0])
    elif kwargs['components_to_plot'] == 'primary':
        x_min = np.min(kwargs['points_primary'][:, 0])
        x_max = np.max(kwargs['points_primary'][:, 0])
    elif kwargs['components_to_plot'] == 'secondary':
        x_min = np.min(kwargs['points_secondary'][:, 0])
        x_max = np.max(kwargs['points_secondary'][:, 0])

    d = (x_max - x_min) / 2
    ax.set_xlim3d(x_min, x_max)
    ax.set_ylim3d(-d, d)
    ax.set_zlim3d(-d, d)
    if kwargs['plot_axis']:
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
    else:
        ax.set_axis_off()
    plt.subplots_adjust(left=0.0, right=1.0, top=1.0, bottom=0.0)

    return fig if kwargs['return_figure_instance'] else plt.show()


def binary_surface_anim(**kwargs):
    """
    Function creates animation of the orbital motion.

    :param kwargs: Dict;
    :**kwargs options**:
        * **start_phase** * -- float;
        * **stop_phase** * -- float;
        * **phase_step** * -- float;
        * **units** * -- units for gravity acceleration colormap
        * **plot_axis** * -- bool, if False, axis will not be displayed
        * **colormap** * -- `temperature`, `gravity_acceleration` or None,
        * **savepath** * -- string or None, animation will be stored to `savepath`
        * **separate_colormaps** * -- bool; if True, figure will contain separate colormap for each component
    """
    def update_plot(frame_number, _points, _faces, _clr, _cmaps, _plot):
        for _, _ in enumerate(_plot):
            p = ax.clear()
        for ii, p in enumerate(_plot):
            ax.set_xlim3d(-kwargs['axis_lim'], kwargs['axis_lim'])
            ax.set_ylim3d(-kwargs['axis_lim'], kwargs['axis_lim'])
            ax.set_zlim3d(-kwargs['axis_lim'], kwargs['axis_lim'])
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')

            p = ax.plot_trisurf(_points[ii][frame_number][:, 0],
                                _points[ii][frame_number][:, 1],
                                _points[ii][frame_number][:, 2],
                                triangles=_faces[ii][frame_number],
                                antialiased=True, shade=False, color=_clr[ii])
            if kwargs.get('colormap', False):
                p.set_cmap(cmap=CMAPS[kwargs.get('colormap', cm.jet)])
                p.set_array(_cmaps[ii][frame_number])

            if kwargs['edges']:
                p.set_edgecolor('black')

            ax.text(-kwargs['axis_lim'], 0.9*kwargs['axis_lim'], 0.8*kwargs['axis_lim'],
                    f"{kwargs['phases'][frame_number]%1.0:.2f}")

            if not kwargs['plot_axis']:
                ax.set_axis_off()

    fig = plt.figure(figsize=(7, 7))
    ax = fig.add_subplot(111, projection='3d')
    ax.set_box_aspect([1, 1, 1])
    ax.elev = 0
    ax.azim = 180

    ax.set_xlim3d(-kwargs['axis_lim'], kwargs['axis_lim'])
    ax.set_ylim3d(-kwargs['axis_lim'], kwargs['axis_lim'])
    ax.set_zlim3d(-kwargs['axis_lim'], kwargs['axis_lim'])

    clr = ['g', 'r']
    cmaps = []

    if not kwargs['separate_colormaps']:
        points = [[up.concatenate((kwargs['points_primary'][ii], kwargs['points_secondary'][ii]), axis=0)
                  for ii in range(kwargs['n_frames'])]]
        faces = [[up.concatenate((kwargs['faces_primary'][ii],
                                      kwargs['faces_secondary'][ii] + np.shape(kwargs['points_primary'][ii])[0]),
                                      axis=0)
                     for ii in range(kwargs['n_frames'])]]

        plot = [ax.plot_trisurf(points[0][0][:, 0], points[0][0][:, 1], points[0][0][:, 2],
                                triangles=faces[0][0], antialiased=True,
                                shade=False, color=clr[0])]
        if kwargs.get('colormap', False):
            plot[0].set_cmap(cmap=CMAPS[kwargs['colormap']])
            cmaps = [[up.concatenate((kwargs['primary_cmap'][ii], kwargs['secondary_cmap'][ii]), axis=0)
                      for ii in range(kwargs['n_frames'])]]
            plot[0].set_array(cmaps[0][0])
    else:
        points = [kwargs['points_primary'], kwargs['points_secondary']]
        faces = [kwargs['faces_primary'], kwargs['faces_secondary']]
        plot = [ax.plot_trisurf(kwargs['points_primary'][0][:, 0], kwargs['points_primary'][0][:, 1],
                                kwargs['points_primary'][0][:, 2], triangles=kwargs['faces_primary'][0],
                                antialiased=True, shade=False, color=clr[0]),
                ax.plot_trisurf(kwargs['points_secondary'][0][:, 0], kwargs['points_secondary'][0][:, 1],
                                kwargs['points_secondary'][0][:, 2], triangles=kwargs['faces_secondary'][0],
                                antialiased=True, shade=False, color=clr[1])]
        if kwargs.get('colormap', False):
            plot[0].set_cmap(cmap=CMAPS[kwargs['colormap']])
            plot[1].set_cmap(cmap=CMAPS[kwargs['colormap']])
            cmaps = [kwargs['primary_cmap'], kwargs['secondary_cmap']]
            plot[0].set_array(cmaps[0][0])
            plot[1].set_array(cmaps[1][0])

    if kwargs.get('edges', False):
        if kwargs['separate_colormaps']:
            plot[0].set_edgecolor('black')
            plot[1].set_edgecolor('black')
        else:
            plot[0].set_edgecolor('black')

    args = (points, faces, clr, cmaps, plot)
    plt.subplots_adjust(left=0.0, right=1.0, top=1.0, bottom=0.0)
    ani = animation.FuncAnimation(fig, update_plot, kwargs['n_frames'], fargs=args, interval=20)
    plt.show() if not kwargs['savepath'] else ani.save(kwargs['savepath'], writer='ffmpeg', fps=30, dpi=300)


def single_surface_anim(**kwargs):
    def update_plot(frame_number, _points, _faces, _clr, _cmaps, _plot):
        for _, _ in enumerate(_plot):
            p = ax.clear()
        for ii, p in enumerate(_plot):
            ax.set_xlim3d(-kwargs['axis_lim'], kwargs['axis_lim'])
            ax.set_ylim3d(-kwargs['axis_lim'], kwargs['axis_lim'])
            ax.set_zlim3d(-kwargs['axis_lim'], kwargs['axis_lim'])
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')

            p = ax.plot_trisurf(_points[ii][frame_number][:, 0],
                                _points[ii][frame_number][:, 1],
                                _points[ii][frame_number][:, 2],
                                triangles=_faces[ii][frame_number],
                                antialiased=True, shade=False, color=_clr[ii])
            if kwargs.get('colormap', False):
                p.set_cmap(cmap=CMAPS[kwargs.get('colormap', cm.jet)])
                p.set_array(_cmaps[ii][frame_number])
                p.set_clim(cmap_limits[0], cmap_limits[1])

            if kwargs['edges']:
                p.set_edgecolor('black')

            ax.text(-kwargs['axis_lim'], 0.9 * kwargs['axis_lim'], 0.8 * kwargs['axis_lim'],
                    f"{kwargs['phases'][frame_number]%1.0:.2f}")

    fig = plt.figure(figsize=(7, 7))
    ax = fig.add_subplot(111, projection='3d')
    ax.set_box_aspect([1, 1, 1])
    ax.elev = 0
    ax.azim = 180

    ax.set_xlim3d(-kwargs['axis_lim'], kwargs['axis_lim'])
    ax.set_ylim3d(-kwargs['axis_lim'], kwargs['axis_lim'])
    ax.set_zlim3d(-kwargs['axis_lim'], kwargs['axis_lim'])

    clr = ['g', 'r']
    cmaps = []
    points = [kwargs['points']]
    faces = [kwargs['faces']]

    plot = [ax.plot_trisurf(points[0][0][:, 0], points[0][0][:, 1], points[0][0][:, 2],
                            triangles=faces[0][0], antialiased=True,
                            shade=False, color=clr[0])]
    if kwargs.get('colormap', False):
        cmap_limits = (np.min(kwargs['cmap']), np.max(kwargs['cmap']))
        plot[0].set_cmap(cmap=CMAPS[kwargs['colormap']])
        cmaps = [kwargs['cmap']]
        plot[0].set_array(cmaps[0][0])

    args = (points, faces, clr, cmaps, plot)
    plt.subplots_adjust(left=0.0, right=1.0, top=1.0, bottom=0.0)
    ani = animation.FuncAnimation(fig, update_plot, kwargs['n_frames'], fargs=args, interval=20)
    plt.show() if not kwargs['savepath'] else ani.save(kwargs['savepath'], writer='ffmpeg', fps=20, dpi=300)


def phase_curve(**kwargs):
    fig = plt.figure(figsize=(8, 6))
    if kwargs['unit'] in ['normalized', 'normalised']:
        C = np.max([kwargs['fluxes'][item] for item in kwargs['fluxes']])
    else:
        C = 1
    for item in kwargs['fluxes']:
        plt.plot(kwargs['phases'], kwargs['fluxes'][item] / C, label=item)
        plt.legend()

    plt.xlabel('Phase')
    if isinstance(kwargs['unit'], type(u.W/u.m**2)):
        uu = kwargs['unit']
        plt.ylabel(f'Flux/({uu:latex})')
    elif isinstance(kwargs['unit'], type(u.mag)):
        plt.ylabel(f'Magnitude')
        plt.gca().invert_yaxis()
    else:
        plt.ylabel('Flux')
    if kwargs['legend']:
        plt.legend(loc=kwargs['legend_location'])

    return fig if kwargs['return_figure_instance'] else plt.show()


def rv_curve(**kwargs):
    fig = plt.figure(figsize=(8, 6))
    phases, rvs = kwargs["phases"], kwargs["rvs"]
    for component in rvs.keys():
        plt.plot(phases, rvs[component], label=component)
    plt.legend()

    plt.xlabel('Phase')
    if isinstance(kwargs['unit'], type(u.m / u.s)):
        uu = kwargs['unit']
        plt.ylabel(f'Radial velocity/({uu:latex})')
    else:
        plt.ylabel('Radial velocity')
    if kwargs['legend']:
        plt.legend(loc=kwargs['legend_location'])

    return fig if kwargs['return_figure_instance'] else plt.show()


def binary_rv_fit_plot(**kwargs):
    """
    Plots the model and residuals described by fit params or calculated by last run of fitting procedure.

    :param kwargs: Dict;
    :**kwargs options**:
        * **fit_params** * -- Dict; {fit_parameter: {value: float, unit: astropy.unit.Unit}
        * **start_phase** * -- float;
        * **stop_phase** * -- float;
        * **number_of_points** * -- int;
        * **y_axis_unit** * -- astropy.unit.Unit;
    :return:
    """
    matplotlib.rcParams.update({'errorbar.capsize': 2})
    fig = plt.figure(figsize=(8, 6))

    gs = gridspec.GridSpec(2, 1, height_ratios=[3, 1])
    ax1 = fig.add_subplot(gs[0])
    ax2 = fig.add_subplot(gs[1], sharex=ax1)

    ax1.plot(kwargs['synth_phases'], kwargs['rv_fit']['primary'], label='primary RV fit', color='cornflowerblue')
    ax1.plot(kwargs['synth_phases'], kwargs['rv_fit']['secondary'], label='secondary RV fit', color='firebrick',
             ls='dashed')
    if kwargs['y_err']['primary'] is None:
        ax1.scatter(kwargs['x_data']['primary'], kwargs['y_data']['primary'],
                    marker='o', color='blue', s=3, label='primary')
        ax2.scatter(kwargs['x_data']['primary'], kwargs['residuals']['primary'],
                    marker='o', color='blue', s=3, label='primary')
    else:
        ax1.errorbar(kwargs['x_data']['primary'], kwargs['y_data']['primary'], yerr=kwargs['y_err']['primary'],
                     linestyle='none', marker='o', color='blue', markersize=3, label='primary')
        ax2.errorbar(kwargs['x_data']['primary'], kwargs['residuals']['primary'], yerr=kwargs['y_err']['primary'],
                     linestyle='none', marker='o', color='blue', markersize=3, label='primary')

    if kwargs['y_err']['secondary'] is None:
        ax1.scatter(kwargs['x_data']['secondary'], kwargs['y_data']['secondary'],
                    marker='x', color='red', s=3, label='secondary')
        ax2.scatter(kwargs['x_data']['secondary'], kwargs['residuals']['secondary'],
                    marker='x', color='red', s=3, label='secondary')
    else:
        ax1.errorbar(kwargs['x_data']['secondary'], kwargs['y_data']['secondary'], yerr=kwargs['y_err']['secondary'],
                     linestyle='none', marker='x', color='red',
                     markersize=3, label='secondary')
        ax2.errorbar(kwargs['x_data']['secondary'], kwargs['residuals']['secondary'], yerr=kwargs['y_err']['secondary'],
                     linestyle='none', marker='x', color='red',
                     markersize=3, label='secondary')
    ax1.legend()
    unit = kwargs['y_unit']
    ax1.set_ylabel(f'Radial velocity/[{unit}]')

    ax2.axhline(0, ls='dashed', c='black', lw=0.5)

    ax2.set_xlabel('Phase')
    ax2.set_ylabel('Residuals')

    plt.subplots_adjust(hspace=0.0, top=0.98, right=0.97)
    return fig if kwargs['return_figure_instance'] else plt.show()


def binary_lc_fit_plot(**kwargs):
    synthetic_clrs = {
        'bolometric': 'black',
        'Generic.Bessell.U': '#ff00bf',
        'Generic.Bessell.B': '#0000E5',
        'Generic.Bessell.V': '#00cc00',
        'Generic.Bessell.R': '#fd2b2b',
        'Generic.Bessell.I': '#b30000',
        'SLOAN.SDSS.u': '#0000ff',
        'SLOAN.SDSS.g': '#00cc00',
        'SLOAN.SDSS.r': '#ff1a1a',
        'SLOAN.SDSS.i': '#cc00cc',
        'SLOAN.SDSS.z': '#00ffff',
        'Generic.Stromgren.u': '#cc00cc',
        'Generic.Stromgren.v': '#ff00ff',
        'Generic.Stromgren.b': '#3333ff',
        'Generic.Stromgren.y': '#00e600',
        'Kepler': '#E50000',
        'GaiaDR2': 'black',
        'TESS': '#003849'
    }
    datapoint_clrs = {
        'bolometric': 'gray',
        'Generic.Bessell.U': '#cc0099',
        'Generic.Bessell.B': '#00007F',
        'Generic.Bessell.V': '#008000',
        'Generic.Bessell.R': '#ff0000',
        'Generic.Bessell.I': '#800000',
        'SLOAN.SDSS.u': '#000099',
        'SLOAN.SDSS.g': '#009900',
        'SLOAN.SDSS.r': '#e60000',
        'SLOAN.SDSS.i': '#800080',
        'SLOAN.SDSS.z': '#00cccc',
        'Generic.Stromgren.u': '#990099',
        'Generic.Stromgren.v': '#cc00cc',
        'Generic.Stromgren.b': '#0000cc',
        'Generic.Stromgren.y': '#00b300',
        'Kepler': '#890000',
        'GaiaDR2': 'gray',
        'TESS': '#006989'
    }

    matplotlib.rcParams.update({'errorbar.capsize': 2})
    fig = plt.figure(figsize=(8, 6))

    gs = gridspec.GridSpec(2, 1, height_ratios=[3, 1])
    ax1 = fig.add_subplot(gs[0])
    ax2 = fig.add_subplot(gs[1], sharex=ax1)

    for fltr, curve in kwargs['lcs'].items():
        rasterize = np.shape(kwargs['x_data'][fltr])[0] > 10000 if kwargs['rasterize'] is None else rasterize
        (dt_clr, clr) = (datapoint_clrs[fltr], datapoint_clrs[fltr]) if len(kwargs['lcs']) > 1 else ('blue', 'red')

        if kwargs['y_err'][fltr] is None:
            ax1.scatter(kwargs['x_data'][fltr], kwargs['y_data'][fltr], s=3, label=fltr + ' observed',
                        color=dt_clr)

            ax2.scatter(kwargs['x_data'][fltr], kwargs['residuals'][fltr], s=3, label=fltr + ' residual',
                        color=dt_clr)
        else:
            ax1.errorbar(kwargs['x_data'][fltr], kwargs['y_data'][fltr], yerr=kwargs['y_err'][fltr],
                         linestyle='none', markersize=3, label=fltr + ' observed', color=dt_clr, rasterized=rasterize)

            ax2.errorbar(kwargs['x_data'][fltr], kwargs['residuals'][fltr], yerr=kwargs['y_err'][fltr],
                         linestyle='none', markersize=3, label=fltr + ' residual', color=dt_clr, rasterized=rasterize)

        ax1.plot(kwargs['synth_phases'], curve, label=fltr + ' synthetic', color=clr, linewidth=2)

    ax2.axhline(0, ls='dashed', c='black', lw=0.5)

    if kwargs['legend']:
        ax1.legend(loc=kwargs['loc'])
    # ax2.legend(loc=1)

    ax1.set_ylabel(f'Normalized flux')

    ax2.set_xlabel('Phase')
    ax2.set_ylabel('Residuals')

    plt.subplots_adjust(hspace=0.0, top=0.98, right=0.97)
    return fig if kwargs['return_figure_instance'] else plt.show()


def passband_plot(**kwargs):
    # Extract values
    xs = kwargs["xs"]
    ys = kwargs["ys"]
    x_unit = kwargs["x_unit"]
    y_unit = kwargs["y_unit"]
    passband = kwargs["passband"]

    # Create plot
    plt.figure(figsize=(8, 6))
    plt.plot(xs, ys, marker="o", linestyle="-", label=passband, lw=0.5, markersize=2)
    plt.xlabel(x_unit)
    plt.ylabel(y_unit)
    plt.legend()
    plt.grid(True)

    plt.show()



import numpy as np

from ... logger import getLogger
from ... import const
from .. container import SinglePositionContainer
from . import utils as crv_utils, c_managed
from ... observer.mp_manager import manage_observations

logger = getLogger('single_system.curves.curves')


def resolve_curve_method(system, fn_array):
    """
    Resolves which curve calculating method to use based on the properties of the SingleSystem.

    :param system: elisa.single_system.SingleSystem;
    :param fn_array: Tuple; list of curve calculating functions in specific order
                            (system with pulsations, system without pulsations)
    :return: curve calculating method chosen from `fn_array`
    """
    if system.star.has_pulsations():
        logger.debug('Calculating light curve for star system with pulsation')
        return fn_array[1]
    else:
        logger.debug('Calculating light curve for a non pulsating single star system')
        return fn_array[0]

    # raise NotImplementedError("System type not implemented or invalid.")


def prep_initial_system(single, **kwargs):
    """
    Prepares base single system from which curves will be calculated in case of single system without pulsations.

    :param single: elisa.single_system.system.SingleSystem;
    :return: elisa.single_system.container.SystemContainer;
    """
    from_this = dict(single_system=single, position=const.Position(0, np.nan, 0.0, np.nan, 0.0))
    initial_system = SinglePositionContainer.from_single_system(**from_this)
    do_pulsations = kwargs.get('build_pulsations', True)
    initial_system.build(do_pulsations)
    return initial_system


def produce_curves_wo_pulsations(single, initial_system, phases, curve_fn, crv_labels, **kwargs):
    """
    General function for creation of single system light curve without pulsations.

    :param single: elisa.single_system.system.SingleSystem;
    :param initial_system: elisa.single_system.container.SystemContainer
    :param phases: numpy.array;
    :param curve_fn: callable; function to calculate given type of the curve
    :param crv_labels: List; labels of the calculated curves (passbands, components,...)
    :param kwargs: Dict;
            * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
            * ** left_bandwidth ** * - float
            * ** right_bandwidth ** * - float
            * ** position_method** * - function definition; to evaluate orbital positions
            * ** phases ** * - numpy.array

    :return: Dict; calculated curves
    """
    crv_utils.prep_surface_params(initial_system, return_values=False, write_to_containers=True, **kwargs)
    fn_args = (single, initial_system, crv_labels, curve_fn)
    return manage_observations(fn=c_managed.produce_curves_wo_pulsations_mp, fn_args=fn_args, position=phases, **kwargs)


def produce_curves_with_pulsations(single, initial_system, phases, curve_fn, crv_labels, **kwargs):
    """
    General function for creation of single system light curve with pulsations.

    :param single: elisa.single_system.system.SingleSystem;
    :param initial_system: elisa.single_system.container.SystemContainer;
    :param phases: numpy.array;
    :param curve_fn: callable; function to calculate given type of the curve
    :param crv_labels: List; labels of the calculated curves (passbands, components,...)
    :param kwargs: Dict;
            * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
            * ** left_bandwidth ** * - float
            * ** right_bandwidth ** * - float
            * ** position_method** * - function definition; to evaluate orbital positions
            * ** phases ** * - numpy.array

    :return: Dict; calculated curves
    """
    fn_args = (single, initial_system, crv_labels, curve_fn)
    return manage_observations(fn=c_managed.produce_curves_with_pulsations_mp, fn_args=fn_args, position=phases, **kwargs)



import numpy as np
from packaging import version

INT = int
FLOAT = float
BOOL = bool
COMPLEX = complex

if version.parse(np.__version__) < version.parse("1.20.0"):
    # noinspection PyUnresolvedReferences
    INT = np.int
    # noinspection PyUnresolvedReferences
    FLOAT = np.float
    # noinspection PyUnresolvedReferences
    BOOL = np.bool
    # noinspection PyUnresolvedReferences
    COMPLEX = np.complex



import os
import json


class Registry(object):
    SCHEMA_SEARCH_DIRS = [
        os.path.join(os.path.dirname(__file__), 'schema_files')
    ]

    @classmethod
    def get_schema(cls, fname):
        """
        Gets schema for given topic.

        :param fname: filename for schema
        :return: Dict; json object
        """
        schema_path = cls._get_schema_path(fname)
        with open(schema_path, "r") as f:
            # return json.loads(f.read())
            schema = json.loads(f.read())

        spot_schema_path = cls._get_schema_path('spot')
        with open(spot_schema_path, "r") as f:
            spot_schema = json.loads(f.read())

        mode_schema_path = cls._get_schema_path('pulsation')
        with open(mode_schema_path, "r") as f:
            mode_schema = json.loads(f.read())

        # adding subschemas for spots and pulsations to the base system schemas
        feature_enabled = ['star', 'primary', 'secondary']
        for component in feature_enabled:
            if component in schema['properties']:
                component_schema = schema['properties'][component]
                if 'spots' in component_schema['properties']:
                    component_schema['properties']['spots'] = spot_schema

                if 'pulsations' in component_schema['properties']:
                    component_schema['properties']['pulsations'] = mode_schema

        return schema

    @classmethod
    def _get_schema_path(cls, fname):
        for base_path in cls.SCHEMA_SEARCH_DIRS:

            path = os.path.join(base_path, f'{fname}.sc')
            if os.path.isfile(path):
                return path
        raise LookupError(f'No schema found for fname: {fname}.')



from ... binary_system.system import BinarySystem
from .. models.serializers import (
    serialize_primary_kwargs,
    serialize_secondary_kwargs,
    serialize_system_kwargs
)


def prepare_binary(discretization=5, _verify=False, **kwargs):
    """
    Setup binary system from initial parameter object.
    If `metallicity` is not supplied 0 is used to initialize the parameter. Similarly, default value of `synchronicity`
    parameter is 1.0.

    :param _verify: bool; verify input json
    :param discretization: float; primary component's surface discretization factor
    :param kwargs: Dict; complete set of model parameters in flat format {'parameter@name': value, }
    :return: elisa.binary_system.system.BinarySystem
    """
    kwargs.update({"primary@discretization_factor": discretization})
    primary_kwargs = serialize_primary_kwargs(**kwargs)
    secondary_kwargs = serialize_secondary_kwargs(**kwargs)
    system_kwargs = serialize_system_kwargs(**kwargs)
    json = {
        "primary": dict(**primary_kwargs),
        "secondary": dict(**secondary_kwargs),
        "system": dict(**system_kwargs)
    }
    return BinarySystem.from_json(json, _verify=_verify)


def synthetic_binary(phases, discretization, observer, **kwargs):
    """
    Function returns synthetic light curve of binary system based on a set of model parameters.

    :param phases: Union[List, numpy.array]; photometric phases in which the curves will be generated
    :param discretization: float; primary component's surface discretization factor
    :param observer: elisa.observer.observer.Observer; observer instance
    :param kwargs: Dict; The structure of the `kwargs` is similar to the input JSON used in BinarySystem.from_json()
                         function, with exception of using the flat format {'parameter@name': value, } instead of the
                         nested structure. The default units for parameters are the default input units
                         (see elisa.units).

    :return: Dict[str, numpy.array]; light curves in in format {'passband_name': LC in filter}
    """
    binary = prepare_binary(discretization, **kwargs)
    observer._system = binary

    lc = observer.observe.lc(phases=phases, normalize=True)
    return lc[1]



import numpy as np
from .. utils import is_empty


def normalize_light_curve(y_data, y_err=None, kind='global_maximum', top_fraction_to_average=0.1):
    """
    Function normalizes light by a method defined by `kind` argument.

    :param y_data: Dict; dictionary containing curves, {filter: np.ndarray, ...}
    :param y_err: Dict; dictionary containing errors, {filter: np.ndarray, ...}
    :param kind: str; specifies kind of normalization
    :**kind options**:
        * **average**- each curve is normalized to its average
        * **global_average** - curves are normalized to their global average
        * **maximum** - each curve is normalized to its own maximum
        * **global_maximum** - curves are normalized to their global maximum
    :param top_fraction_to_average: float; top portion of the dataset (in y-axis direction) used in the
                                           normalization process, from (0, 1) interval
    :return: Dict;
    """
    valid_arguments = ['average', 'global_average', 'maximum', 'global_maximum', 'minimum']
    if kind == valid_arguments[0]:  # each curve is normalized to its average
        coeff = {key: np.mean(val) for key, val in y_data.items()}
    elif kind == valid_arguments[1]:  # curves are normalized to their global average
        c = np.mean(list(y_data.values()))
        coeff = {key: c for key, val in y_data.items()}
    elif kind == valid_arguments[2]:  # each curve is normalized to its own average of the top fraction
        n = {key: int(top_fraction_to_average * len(val)) + 1 for key, val in y_data.items()}
        coeff = {key: np.average(val[np.argsort(val)[-n[key]:]]) for key, val in y_data.items()}
    elif kind == valid_arguments[3]:  # curves are normalized to their global maximum
        vals = np.concatenate(list(y_data.values()))
        n = int(top_fraction_to_average * len(vals) / len(y_data)) + 1
        c = np.average(vals[np.argsort(vals)[-n:]])
        coeff = {key: c for key, val in y_data.items()}
    elif kind == valid_arguments[4]:  # each curve is normalized to its own average of the top fraction
        n = {key: int(top_fraction_to_average * len(val)) for key, val in y_data.items()}
        coeff = {key: np.average(val[np.argsort(val)[:n[key]]]) for key, val in y_data.items()}
    else:
        raise ValueError(f'Argument `kind` = {kind} is not one of the valid arguments {valid_arguments}')

    y_data = {key: np.array(val) / coeff[key] for key, val in y_data.items()}
    y_err = {key: np.array(val) / coeff[key] if not is_empty(val) else None for key, val in y_err.items()} \
        if not is_empty(y_err) else None
    return y_data, y_err


def adjust_flux_for_distance(curves, distance):
    """
    Adjusting the light curve flux to levels corresponding to the system`s distance.

    :param curves: dict; band-wise flux
    :param distance: float; distance to the observer
    :return: dict; corrected band-wise flux
    """
    d_squared = np.power(distance, 2)
    return {band: curve/d_squared for band, curve in curves.items()}


def convert_to_magnitudes(curves, zero_points):
    """
    Conversion from flux to magnitudes.

    :param curves: dict;
    :param zero_points: dict; fluxes corresponding to 0 mag
    :return: dict;
    """
    ret_dict = dict()
    for band, curve in curves.items():
        if zero_points['reference_magnitudes'][band] is None:
            raise ValueError(f'Calibration reference magnitude is not available for the filter {band}')
        ret_dict[band] = zero_points['reference_magnitudes'][band]-2.5 * np.log10(curve/zero_points['fluxes'][band])

    return ret_dict



import json
import os
import warnings
import numpy as np
import os.path as op

from configparser import ConfigParser
from logging import config as log_conf
from os.path import dirname, isdir

from .. schema_registry import registry
from .. managers.settings_manager import SettingsManager, DefaultSettings


first_time_user = False
c_parse = ConfigParser()
env_variable_config = os.environ.get('ELISA_CONFIG', '')
venv_config = op.join(os.environ.get('VIRTUAL_ENV', ''), 'conf', 'elisa_conf.ini')
default_config = op.join(DefaultSettings.HOME, "config.ini")

# read configuration file
if op.isfile(env_variable_config):
    config_file = env_variable_config
elif op.isfile(venv_config):
    config_file = venv_config
elif op.isfile(default_config):
    config_file = default_config
else:
    # provide a minimal configuration for first time running users
    config_file = default_config
    first_time_user = True
    SettingsManager.run()


class _Const(object):
    # PASSBAND RELATED *************************************************************************************************
    PASSBANDS = [
        'bolometric',
        'Generic.Bessell.U',
        'Generic.Bessell.B',
        'Generic.Bessell.V',
        'Generic.Bessell.R',
        'Generic.Bessell.I',
        'SLOAN.SDSS.u',
        'SLOAN.SDSS.g',
        'SLOAN.SDSS.r',
        'SLOAN.SDSS.i',
        'SLOAN.SDSS.z',
        'Generic.Stromgren.u',
        'Generic.Stromgren.v',
        'Generic.Stromgren.b',
        'Generic.Stromgren.y',
        'Kepler',
        # 'GaiaDR2',
        'TESS',
        'Gaia.2010.G',
        'Gaia.2010.BP',
        'Gaia.2010.RP',
    ]

    PASSBAND_DATAFRAME_THROUGHPUT = "throughput"
    PASSBAND_DATAFRAME_WAVE = "wavelength"
    # ******************************************************************************************************************

    # ATM RELATED ******************************************************************************************************
    ATM_MODEL_DATAFRAME_FLUX = "flux"
    ATM_MODEL_DATAFRAME_WAVE = "wave"

    ATM_MODEL_DATAFARME_DTYPES = {
        ATM_MODEL_DATAFRAME_FLUX: float,
        ATM_MODEL_DATAFRAME_WAVE: float
    }
    # ******************************************************************************************************************

    # LIMB DARKENING RELATED *******************************************************************************************
    LD_LAW_TO_FILE_PREFIX = {
        "linear": "lin",
        "cosine": "lin",
        "logarithmic": "log",
        "square_root": "sqrt",
    }

    AVAILABLE_LD_LAWS = list(LD_LAW_TO_FILE_PREFIX.keys())

    LD_LAW_CFS_COLUMNS = {
        "linear": ["xlin"],
        "cosine": ["xlin"],
        "logarithmic": ["xlog", "ylog"],
        "square_root": ["xsqrt", "ysqrt"],
    }

    BASIC_COLS = ["temperature", "gravity"]
    LD_LAW_COLS_ORDER = {
        "linear": BASIC_COLS + LD_LAW_CFS_COLUMNS["linear"],
        "cosine": BASIC_COLS + LD_LAW_CFS_COLUMNS["cosine"],
        "logarithmic": BASIC_COLS + LD_LAW_CFS_COLUMNS["logarithmic"],
        "square_root": BASIC_COLS + LD_LAW_CFS_COLUMNS["square_root"]
    }

    LD_DOMAIN_COLS = ["temperature", "gravity"]
    # ******************************************************************************************************************

    BINARY_COUNTERPARTS = {"primary": "secondary", "secondary": "primary"}

    ATLAS_TO_ATM_FILE_PREFIX = {
        "castelli": "ck",
        "castelli-kurucz": "ck",
        "ck": "ck",
        "ck04": "ck",
        "kurucz": "k",
        "k": "k",
        "k93": "k"
    }

    ATM_ATLAS_NORMALIZER = {
        **ATLAS_TO_ATM_FILE_PREFIX,
        **{"bb": "bb", "black_body": "bb"}
    }

    ATM_DOMAIN_QUANTITY_TO_VARIABLE_SUFFIX = {
        "temperature": "TEMPERATURE_LIST_ATM",
        "gravity": "GRAVITY_LIST_ATM",
        "metallicity": "METALLICITY_LIST_ATM"
    }

    DATETIME_MASK = '%Y-%m-%dT%H.%M.%S'
    DATE_MASK = '%Y-%m-%d'

    DATASET_MANDATORY_KWARGS = ['x_data', 'y_data', 'x_unit', 'y_unit']
    DATASET_OPTIONAL_KWARGS = ['y_err']

    DELIM_WHITESPACE = r'\s+|\t+|\s+\t+|\t+\s+'
    DATA_PATH = op.join(dirname(dirname(__file__)), "data")

    # paths to mesh correction factors
    PATH_TO_SINGLE_CORRECTIONS = op.join(DATA_PATH, 'mesh_corrections', 'correction_factors_single.npy')
    PATH_TO_DETACHED_CORRECTIONS = op.join(DATA_PATH, 'mesh_corrections', 'correction_factors_detached.npy')
    PATH_TO_OVER_CONTACT_CORRECTIONS = op.join(
        DATA_PATH, 'mesh_corrections', 'correction_factors_over-contact.npy'
    )
    PATH_TO_ALBEDOS = op.join(DATA_PATH, 'albedos.json')
    PATH_TO_BETA = op.join(DATA_PATH, 'gravity_darkening.json')


class Settings(_Const, DefaultSettings):
    _instance = None

    # defaults #########################################################################################################
    DEFAULT_SETTINGS = {}

    # schema registry
    SCHEMA_REGISTRY = registry.Registry()

    # basic app configuration
    CONFIG_FILE = config_file
    FIRST_TIME_USER = first_time_user
    ####################################################################################################################

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Settings, cls).__new__(cls)
            cls.DEFAULT_SETTINGS = cls.settings_serializer()
            # Put any initialization here.
            cls.read_and_update_config()
            cls._update_atlas_to_base_dir()
            cls._create_home()

        return cls._instance

    @classmethod
    def settings_serializer(cls):
        return {
            "CONFIG_FILE": cls.CONFIG_FILE,
            "LOG_CONFIG": cls.LOG_CONFIG,
            "SUPPRESS_WARNINGS": cls.SUPPRESS_WARNINGS,
            "SUPPRESS_LOGGER": cls.SUPPRESS_LOGGER,
            "HOME": cls.HOME,
            "REFLECTION_EFFECT": cls.REFLECTION_EFFECT,
            "REFLECTION_EFFECT_ITERATIONS": cls.REFLECTION_EFFECT_ITERATIONS,
            "LIMB_DARKENING_LAW": cls.LIMB_DARKENING_LAW,
            "DEFAULT_TEMPERATURE_PERTURBATION_PHASE_SHIFT": cls.DEFAULT_TEMPERATURE_PERTURBATION_PHASE_SHIFT,
            "SURFACE_DISPLACEMENT_TOL": cls.SURFACE_DISPLACEMENT_TOL,
            "RV_METHOD": cls.RV_METHOD,
            "RV_LAMBDA_INTERVAL": cls.RV_LAMBDA_INTERVAL,
            "MAX_DISCRETIZATION_FACTOR": cls.MAX_DISCRETIZATION_FACTOR,
            "MIN_DISCRETIZATION_FACTOR": cls.MIN_DISCRETIZATION_FACTOR,
            "DEFAULT_DISCRETIZATION_FACTOR": cls.DEFAULT_DISCRETIZATION_FACTOR,
            "NUMBER_OF_THREADS": cls.NUMBER_OF_THREADS,
            "NUMBER_OF_PROCESSES": cls.NUMBER_OF_PROCESSES,
            "NUMBER_OF_MCMC_PROCESSES": cls.NUMBER_OF_MCMC_PROCESSES,
            "MAX_NU_SEPARATION": cls.MAX_NU_SEPARATION,
            "MAX_D_FLUX": cls.MAX_D_FLUX,
            "MAX_SPOT_D_LONGITUDE": cls.MAX_SPOT_D_LONGITUDE,
            "MAX_SOLVER_ITERS": cls.MAX_SOLVER_ITERS,
            "MAX_CURVE_DATA_POINTS": cls.MAX_CURVE_DATA_POINTS,
            "MIN_POINTS_IN_ECLIPSE": cls.MIN_POINTS_IN_ECLIPSE,
            "TIMER": cls.TIMER,
            "PASSBAND_TABLES": cls.PASSBAND_TABLES,
            "LD_TABLES": cls.LD_TABLES,
            "CK04_ATM_TABLES": cls.CK04_ATM_TABLES,
            "K93_ATM_TABLES": cls.K93_ATM_TABLES,
            "ATM_ATLAS": cls.ATM_ATLAS,
            "MESH_GENERATOR": cls.MESH_GENERATOR,
            "DEFORMATION_TOL": cls.DEFORMATION_TOL,
            "PULSATION_MODEL": cls.PULSATION_MODEL,
            "MCMC_SAVE_INTERVAL": cls.MCMC_SAVE_INTERVAL,
            "CUDA": cls.CUDA,
            "USE_SINGLE_LD_COEFFICIENTS": cls.USE_SINGLE_LD_COEFFICIENTS,
            "USE_INTERPOLATION_APPROXIMATION": cls.USE_INTERPOLATION_APPROXIMATION,
            "USE_SYMMETRICAL_COUNTERPARTS_APPROXIMATION": cls.USE_SYMMETRICAL_COUNTERPARTS_APPROXIMATION,
            "USE_SIMILAR_NEIGHBOURS_APPROXIMATION": cls.USE_SIMILAR_NEIGHBOURS_APPROXIMATION,
            "MAGNITUDE_SYSTEM": cls.MAGNITUDE_SYSTEM
        }

    @staticmethod
    def load_conf(path):
        with open(path) as f:
            conf_dict = json.loads(f.read())
        log_conf.dictConfig(conf_dict)

    @classmethod
    def set_up_logging(cls):
        if op.isfile(cls.LOG_CONFIG):
            cls.load_conf(cls.LOG_CONFIG)
            return
        elif cls.LOG_CONFIG == 'default':
            cls.LOG_CONFIG = op.join(dirname(op.abspath(__file__)), 'logging_schemas/default.json')
        elif cls.LOG_CONFIG == 'fit':
            cls.LOG_CONFIG = op.join(dirname(op.abspath(__file__)), 'logging_schemas/fit.json')
        else:
            cls.LOG_CONFIG = op.join(dirname(op.abspath(__file__)), 'logging_schemas/default.json')

        cls.load_conf(cls.LOG_CONFIG)

    @classmethod
    def read_and_update_config(cls, conf_path=None):
        if not conf_path:
            conf_path = cls.CONFIG_FILE

        if not op.isfile(conf_path):
            msg = (
                "Couldn't find configuration file. Using default settings.\n"
                "   To customize configuration using file either\n"
                "    - specify config with environment variable ELISA_CONFIG\n"
                "    - add conf/elisa_conf.ini under your virtualenv root \n")
            warnings.warn(msg, Warning)
            return

        c_parse.read(conf_path)
        cls.update_config()
        cls._update_atlas_to_base_dir()
        cls._create_home()

    @classmethod
    def configure(cls, **kwargs):
        for key, value in kwargs.items():
            if not hasattr(cls, key):
                raise ValueError("You are about to set configuration which doesn't exist")
            setattr(cls, key, value)
            if key == 'LOG_CONFIG':
                cls.set_up_logging()
        cls._update_atlas_to_base_dir()

    @classmethod
    def _create_home(cls):
        os.makedirs(cls.HOME, exist_ok=True)

    @classmethod
    def update_config(cls):
        cls.CUDA = False
        # **************************************************************************************************************
        if c_parse.has_section('general'):
            cls.SUPPRESS_WARNINGS = c_parse.getboolean('general', 'suppress_warnings', fallback=cls.SUPPRESS_WARNINGS)
            cls.LOG_CONFIG = c_parse.get('general', 'log_config', fallback=cls.LOG_CONFIG)

            cls.set_up_logging()

            cls.SUPPRESS_LOGGER = c_parse.getboolean('general', 'suppress_logger', fallback=cls.SUPPRESS_LOGGER)
            cls.HOME = c_parse.get('general', 'home', fallback=cls.HOME)
        # **************************************************************************************************************
        if c_parse.has_section('physics'):
            cls.REFLECTION_EFFECT = c_parse.getboolean('physics', 'reflection_effect', fallback=cls.REFLECTION_EFFECT)
            cls.REFLECTION_EFFECT_ITERATIONS = c_parse.getint('physics', 'reflection_effect_iterations',
                                                              fallback=cls.REFLECTION_EFFECT_ITERATIONS)
            cls.LIMB_DARKENING_LAW = c_parse.get('physics', 'limb_darkening_law', fallback=cls.LIMB_DARKENING_LAW)
            if cls.LIMB_DARKENING_LAW not in ['linear', 'cosine', 'logarithmic', 'square_root']:
                raise ValueError(f'{cls.LIMB_DARKENING_LAW} is not valid name of limb darkening law. '
                                 f'Available limb darkening laws are: `linear` '
                                 f'or `cosine`, `logarithmic`, `square_root`')

            cls.DEFAULT_TEMPERATURE_PERTURBATION_PHASE_SHIFT = \
                c_parse.getfloat('physics', 'default_temperature_perturbation_phase_shift',
                                 fallback=cls.DEFAULT_TEMPERATURE_PERTURBATION_PHASE_SHIFT)
            cls.SURFACE_DISPLACEMENT_TOL = \
                c_parse.getfloat('physics', 'surface_displacement_tol', fallback=cls.SURFACE_DISPLACEMENT_TOL)
            cls.RV_METHOD = c_parse.getfloat('physics', 'rv_method', fallback=cls.RV_METHOD)
            cls.RV_LAMBDA_INTERVAL = c_parse.getfloat('physics', 'rv_lambda_interval', fallback=cls.RV_LAMBDA_INTERVAL)
            cls.PULSATION_MODEL = c_parse.getfloat('physics', 'pulsation_model', fallback=cls.PULSATION_MODEL)
        # **************************************************************************************************************
        if c_parse.has_section('computational'):
            cls.MAX_DISCRETIZATION_FACTOR = c_parse.getfloat('computational', 'max_discretization_factor',
                                                             fallback=cls.MAX_DISCRETIZATION_FACTOR)
            cls.MIN_DISCRETIZATION_FACTOR = c_parse.getfloat('computational', 'min_discretization_factor',
                                                             fallback=cls.MIN_DISCRETIZATION_FACTOR)
            cls.DEFAULT_DISCRETIZATION_FACTOR = c_parse.getint(
                'computational', 'default_discretization_factor', fallback=cls.DEFAULT_DISCRETIZATION_FACTOR
            )
            cls.NUMBER_OF_THREADS = c_parse.getint('computational', 'number_of_threads', fallback=cls.NUMBER_OF_THREADS)
            if cls.NUMBER_OF_THREADS <= 0:
                raise ValueError("Invalid value for `number_of_threads`, allowed >= 1")

            cls.NUMBER_OF_PROCESSES = c_parse.getint('computational', 'number_of_processes',
                                                     fallback=cls.NUMBER_OF_PROCESSES)
            if cls.NUMBER_OF_PROCESSES > os.cpu_count():
                if not cls.SUPPRESS_WARNINGS:
                    warnings.warn("argument number_of_processes is too big, fallback to number of machine cores")
                cls.NUMBER_OF_PROCESSES = int(os.cpu_count())

            cls.NUMBER_OF_MCMC_PROCESSES = c_parse.getint('computational', 'number_of_mcmc_processes',
                                                          fallback=cls.NUMBER_OF_MCMC_PROCESSES)
            if cls.NUMBER_OF_MCMC_PROCESSES > os.cpu_count():
                if not cls.SUPPRESS_WARNINGS:
                    warnings.warn("argument number_of_mcmc_processes is too big, fallback to number of machine cores")
                cls.NUMBER_OF_MCMC_PROCESSES = int(os.cpu_count())

            cls.MAX_NU_SEPARATION = c_parse.getfloat('computational', 'max_nu_separation',
                                                     fallback=cls.MAX_NU_SEPARATION)
            cls.MAX_D_FLUX = c_parse.getfloat('computational', 'max_d_flux', fallback=cls.MAX_D_FLUX)
            cls.MAX_SPOT_D_LONGITUDE = c_parse.getfloat('computational', 'max_spot_d_longitude',
                                                        fallback=cls.MAX_SPOT_D_LONGITUDE)
            cls.MAX_SOLVER_ITERS = c_parse.getfloat('computational', 'max_solver_iters', fallback=cls.MAX_SOLVER_ITERS)
            cls.MAX_CURVE_DATA_POINTS = c_parse.getfloat('computational', 'max_curve_datapoints',
                                                         fallback=cls.MAX_CURVE_DATA_POINTS)
            cls.MIN_POINTS_IN_ECLIPSE = c_parse.getint('computational', 'min_points_in_eclipse',
                                                       fallback=cls.MIN_POINTS_IN_ECLIPSE)
            cls.MESH_GENERATOR = c_parse.getfloat('computational', 'mesh_generator', fallback=cls.MESH_GENERATOR)
            cls.DEFORMATION_TOL = c_parse.getfloat('computational', 'deformation_tol', fallback=cls.DEFORMATION_TOL)
            cls.MCMC_SAVE_INTERVAL = c_parse.getfloat('computational', 'mcmc_save_interval',
                                                      fallback=cls.MCMC_SAVE_INTERVAL)

            cls.CUDA = c_parse.getboolean('computational', 'cuda', fallback=cls.CUDA)

            if cls.CUDA:
                try:
                    # noinspection PyUnresolvedReferences
                    import cumpy
                    if not cumpy.cuda.is_available():
                        cls.CUDA = False
                        warnings.warn("You have no CUDA enabled/available on your device. "
                                      "Runtime continue with CPU.", UserWarning)

                except ImportError:
                    warnings.warn("You need to install `pytorch` with cuda to be "
                                  "able to use CUDA features. Fallback to CPU.", UserWarning)
                    cls.CUDA = False

            cls.USE_SINGLE_LD_COEFFICIENTS = c_parse.getboolean('computational', 'use_single_ld_coefficients',
                                                                fallback=cls.USE_SINGLE_LD_COEFFICIENTS)

            cls.USE_INTERPOLATION_APPROXIMATION = c_parse.getboolean(
                'computational', 'use_interpolation_approximation', fallback=cls.USE_INTERPOLATION_APPROXIMATION
            )

            cls.USE_SYMMETRICAL_COUNTERPARTS_APPROXIMATION = c_parse.getboolean(
                'computational', 'use_symmetrical_counterparts_approximation',
                fallback=cls.USE_SYMMETRICAL_COUNTERPARTS_APPROXIMATION
            )

            cls.USE_SIMILAR_NEIGHBOURS_APPROXIMATION = c_parse.getboolean(
                'computational', 'use_similar_neighbours_approximation',
                fallback=cls.USE_SIMILAR_NEIGHBOURS_APPROXIMATION
            )

        # **************************************************************************************************************
        if c_parse.has_section('support'):
            cls.LD_TABLES = c_parse.get('support', 'ld_tables', fallback=cls.LD_TABLES)

            if not isdir(cls.LD_TABLES) and not cls.SUPPRESS_WARNINGS:
                warnings.warn(f"path {cls.LD_TABLES} to limb darkening tables doesn't exists\n"
                              f"Specifiy it in elisa_conf.ini file")

            cls.CK04_ATM_TABLES = c_parse.get('support', 'castelli_kurucz_04_atm_tables', fallback=cls.CK04_ATM_TABLES)

            if not op.isdir(cls.CK04_ATM_TABLES) and not cls.SUPPRESS_WARNINGS:
                warnings.warn(f"path {cls.CK04_ATM_TABLES}\n"
                              f"to castelli-kurucz 2004 atmosphere atlas doesn't exists\n"
                              f"Specifiy it in elisa_conf.ini file")

            cls.K93_ATM_TABLES = c_parse.get('support', 'kurucz_93_atm_tables', fallback=cls.K93_ATM_TABLES)

            if not op.isdir(cls.K93_ATM_TABLES) and not cls.SUPPRESS_WARNINGS:
                warnings.warn(f"path {cls.K93_ATM_TABLES}\n"
                              "to kurucz 1993 atmosphere atlas doesn't exists\n"
                              "Specifiy it in elisa_conf.ini file", UserWarning)

            if c_parse.get('support', 'atlas', fallback=None):
                with np.errstate(divide='ignore', invalid='ignore'):
                    if not cls.SUPPRESS_WARNINGS:
                        warnings.simplefilter("always", DeprecationWarning)
                        warnings.warn("Variable `atlas` in configuration section `support` is not "
                                      "longer supported and will be removed in future version.\n"
                                      "Use atmosphere definition as initial parameter "
                                      "for given celestial object", DeprecationWarning)
                        warnings.simplefilter("ignore", DeprecationWarning)
            cls.ATM_ATLAS = c_parse.get('support', 'atlas', fallback=cls.ATM_ATLAS)
            cls.PASSBAND_TABLES = c_parse.get('support', 'passband_tables', fallback=cls.PASSBAND_TABLES)

            if not isdir(cls.PASSBAND_TABLES) and not cls.SUPPRESS_WARNINGS:
                warnings.warn(f"path {cls.PASSBAND_TABLES} to passband tables doesn't exists\n"
                              f"Specifiy it in elisa_conf.ini file", UserWarning)

            cls.MAGNITUDE_SYSTEM = c_parse.get('support', 'magnitude_system', fallback='vega')
        # **************************************************************************************************************

    @classmethod
    def _update_atlas_to_base_dir(cls):
        cls.ATLAS_TO_BASE_DIR.update({
            "castelli": cls.CK04_ATM_TABLES,
            "castelli-kurucz": cls.CK04_ATM_TABLES,
            "ck": cls.CK04_ATM_TABLES,
            "ck04": cls.CK04_ATM_TABLES,
            "kurucz": cls.K93_ATM_TABLES,
            "k": cls.K93_ATM_TABLES,
            "k93": cls.K93_ATM_TABLES
        })


settings = Settings()



import functools
import numpy as np

from typing import Dict
from scipy import interpolate
from abc import ABCMeta
from scipy.optimize import least_squares

from . shared import (
    AbstractRVFit, AbstractLCFit,
    lc_r_squared, rv_r_squared, r_squared
)

from .. import RVData, LCData
from .. models import rv as rv_model
from .. models import lc as lc_model
from .. models import cost_fns
from .. tools.utils import time_layer_resolver
from .. params import parameters
from ... observer.utils import normalize_light_curve
from ... logger import getPersistentLogger
from ... import const
from ... import settings
from ... binary_system.system import BinarySystem
from ... binary_system.curves.community import RadialVelocitySystem

logger = getPersistentLogger('analytics.binary_fit.least_squares')


def logger_decorator(suppress_logger=False):
    def do(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not suppress_logger:
                logger.debug(f'current xn value: {kwargs}')
            return func(*args, **kwargs)
        return wrapper
    return do


class LightCurveFit(AbstractLCFit, metaclass=ABCMeta):
    """
    General class for solving inverse problem in case of LC data.
    """
    MORPHOLOGY = None

    def model_to_fit(self, xn):
        """
        Cost function minimized during solution of the inverse problem using the Least Squares method.

        :param xn: Iterable[float]; vector containing normalized values of model parameters optimized during fit
        :return: float; error weighted sum of squares of residuals
        """
        diff = 1.0 / self.interp_treshold

        xn = parameters.vector_renormalizer(xn, self.fitable.keys(), self.normalization)
        kwargs = parameters.prepare_properties_set(xn, self.fitable.keys(), self.constrained, self.fixed)
        phases, kwargs = time_layer_resolver(self.x_data_reduced, pop=False, **kwargs)

        if self.fit_xs is None:
            fit_xs = np.linspace(np.min(phases) - diff, np.max(phases) + diff, num=self.interp_treshold + 2) \
                if np.shape(phases)[0] > self.interp_treshold else phases
        else:
            # noinspection PyUnresolvedReferences
            fit_xs = self.fit_xs
        args = fit_xs, self.discretization, self.observer

        kwargs = parameters.extend_json_with_atm_params(
            kwargs,
            atmosphere_model=self.atmosphere_model,
            limb_darkening_coefficients=self.limb_darkening_coefficients
        )

        fn = lc_model.synthetic_binary

        try:
            synthetic = logger_decorator()(fn)(*args, **kwargs)
        except Exception as e:
            logger.error(f'your initial parameters lead during fitting to invalid binary system, exectpion: {str(e)}')
            return const.MAX_USABLE_FLOAT

        if np.shape(fit_xs) != np.shape(phases):
            synthetic = {
                band: interpolate.interp1d(fit_xs, curve, kind='cubic')(phases[self.x_data_reducer[band]])
                for band, curve in synthetic.items()
            }
        else:
            synthetic = {band: val[self.x_data_reducer[band]] for band, val in synthetic.items()}

        synthetic, _ = normalize_light_curve(synthetic, kind='average')

        residuals = cost_fns.wssr(self.y_data, self.y_err, synthetic)

        logger.info(f'current R2: {r_squared(synthetic, self.y_data)}')

        return residuals

    def fit(self, data: Dict[str, LCData], x0: parameters.BinaryInitialParameters,
            discretization=5.0, interp_treshold=None, samples="uniform", **kwargs):
        """
        Fit method using non-linear least squares.
        Based on `https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html`.

        :param data: Dict[elisa.analytics.dataset.base.LCData]; observational data in photometric filters
        :param x0: BinaryInitialParameters; initial state of model parameters
        :param discretization: float; discretization factor used for the primary component
        :param interp_treshold: int; Above this total number of datapoints, light curve will be interpolated
                                     using model containing `interp_treshold` equidistant points per epoch
        :param samples: Union[str, List]; 'uniform' (equidistant sampling in phase), 'adaptive'
                                          (equidistant sampling on curve) or list with phases in (0, 1) interval
        :param kwargs: optional arguments for least_squares function (see documentation for
                       scipy.optimize.least_squares method)
        :return: Dict; optimized model parameters in standard JSON format
        """
        self.set_up(x0, data, passband=data.keys(), discretization=discretization, morphology=self.MORPHOLOGY,
                    interp_treshold=settings.MAX_CURVE_DATA_POINTS if interp_treshold is None else interp_treshold,
                    observer_system_cls=BinarySystem, samples=samples)
        initial_vector = parameters.vector_normalizer(self.initial_vector, self.fitable.keys(), self.normalization)

        # evaluate least squares from scipy
        logger.info("fitting started...")
        result = least_squares(self.model_to_fit, initial_vector, jac=kwargs.get('jac', '2-point'), bounds=(0, 1),
                               method=kwargs.get('method', 'trf'), ftol=kwargs.get('ftol', 1e-7),
                               xtol=kwargs.get('xtol', 1e-8), gtol=kwargs.get('gtol', 1e-8),
                               x_scale=kwargs.get('x_scale', 1.0), loss=kwargs.get('loss', 'linear'),
                               f_scale=kwargs.get('f_scale', 1.0), diff_step=kwargs.get('diff_step', None),
                               tr_solver=kwargs.get('tr_solver', None), tr_options=kwargs.get('tr_options', {}),
                               jac_sparsity=kwargs.get('jac_sparsity', None), max_nfev=kwargs.get('max_nfev', None),
                               verbose=kwargs.get('verbose', 2), args=kwargs.get('args', ()),
                               kwargs=kwargs.get('kwargs', {}))
        logger.info("fitting finished")

        result = parameters.vector_renormalizer(result.x, self.fitable.keys(), self.normalization)
        # put all together
        result_dict = {lbl: {
            'value': result[i],
            'fixed': False,
            'unit': self.fitable[lbl].to_dict()['unit'],
            'min': self.fitable[lbl].min,
            'max': self.fitable[lbl].max,
        } for i, lbl in enumerate(self.fitable.keys())}

        result_dict.update({lbl: {
            "value": val.value,
            'fixed': True,
            'unit': val.to_dict()['unit']
        } for lbl, val in self.fixed.items()})
        result_dict = self.eval_constrained_results(result_dict, self.constrained)

        r_squared_args = (self.x_data_reduced, self.y_data, self.observer.passband, discretization,
                          self.x_data_reducer, 1.0 / self.interp_treshold, self.interp_treshold,
                          self.observer.system_cls)

        r_dict = {key: value['value'] for key, value in result_dict.items()}
        r_dict = parameters.extend_json_with_atm_params(
            r_dict,
            atmosphere_model=self.atmosphere_model,
            limb_darkening_coefficients=self.limb_darkening_coefficients
        )

        r_squared_result = lc_r_squared(lc_model.synthetic_binary, *r_squared_args, **r_dict)
        result_dict["r_squared"] = {'value': r_squared_result, "unit": None}

        setattr(self, 'flat_result', result_dict)
        return parameters.serialize_result(result_dict)


class OvercontactLightCurveFit(LightCurveFit):
    """
    Optimization class for solving an inverse problem for overcontact systems.
    """
    MORPHOLOGY = 'over-contact'


class DetachedLightCurveFit(LightCurveFit):
    """
    Optimization class for solving an inverse problem for detached systems.
    """
    MORPHOLOGY = 'detached'


class CentralRadialVelocity(AbstractRVFit):
    """
    Class for fitting radial velocities using kinematic method.
    """
    def prepare_synthetic(self, xn):
        """
        Returns synthetic radial velocity observations for given set of normalized variable parameters.

        :param xn: List[float]; variable model parameters in normalized form
        :return: Dict[numpy.array]; synthetic RV observations
        """
        xn = parameters.vector_renormalizer(xn, self.fitable.keys(), self.normalization)
        kwargs = parameters.prepare_properties_set(xn, self.fitable.keys(), self.constrained, self.fixed)
        fn = rv_model.central_rv_synthetic
        synthetic = logger_decorator()(fn)(self.x_data_reduced, self.observer, **kwargs)
        return synthetic

    def central_rv_model_to_fit(self, xn):
        """
        Cost function minimized during solution of the inverse problem using the Least Squares method.

        :param xn: List[float]; variable model parameters in normalized form
        :return: numpy.array; error weighted sum of squares of residuals
        """
        synthetic = self.prepare_synthetic(xn)
        synthetic = {comp: synthetic[comp][self.x_data_reducer[comp]] for comp in synthetic}
        return cost_fns.wssr(self.y_data, self.y_err, synthetic)

    def fit(self, data: Dict[str, RVData], x0: parameters.BinaryInitialParameters, **kwargs):
        """
        Method to provide fitting of radial velocities curves.
        It can handle standard physical parameters including component masses `M_1`, `M_2` or astro community
        parameters containing `asini` and `q`. Optimizer based on non-linear least squares method:
        [https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html]

        :param data: elisa.analytics.dataset.base.RVData; radial velocity observations for primary and secondary
                                                          component
        :param x0: elisa.analytics.params.parameters.BinaryInitialParameters; initial state of model parameters
        :param kwargs: optional arguments for least_squares function (see documentation for
                       scipy.optimize.least_squares method)
        :return: Dict; optimized model parameters in standard JSON format
        """
        self.set_up(x0, data, observer_system_cls=RadialVelocitySystem)
        logger.info("fitting radial velocity light curve...")
        func = self.central_rv_model_to_fit
        initial_vector = parameters.vector_normalizer(self.initial_vector, self.fitable.keys(), self.normalization)
        result = least_squares(func, initial_vector, jac=kwargs.get('jac', '2-point'), bounds=(0, 1),
                               method=kwargs.get('method', 'trf'), ftol=kwargs.get('ftol', 1e-8),
                               xtol=kwargs.get('xtol', 1e-8), gtol=kwargs.get('gtol', 1e-8),
                               x_scale=kwargs.get('x_scale', 1.0), loss=kwargs.get('loss', 'linear'),
                               f_scale=kwargs.get('f_scale', 1.0), diff_step=kwargs.get('diff_step', None),
                               tr_solver=kwargs.get('tr_solver', None), tr_options=kwargs.get('tr_options', {}),
                               jac_sparsity=kwargs.get('jac_sparsity', None), max_nfev=kwargs.get('max_nfev', None),
                               verbose=kwargs.get('verbose', 0), args=kwargs.get('args', ()),
                               kwargs=kwargs.get('kwargs', {}))
        logger.info("fitting finished...")
        result = parameters.vector_renormalizer(result.x, self.fitable.keys(), self.normalization)

        # this relies on dict ordering (python >= 3.6)
        result_dict = {lbl: {
            'value': result[i],
            'fixed': False,
            'unit': self.fitable[lbl].to_dict()['unit'],
            'min': self.fitable[lbl].min,
            'max': self.fitable[lbl].max,
        } for i, lbl in enumerate(self.fitable.keys())}

        result_dict.update({lbl: {
            "value": val.value,
            'fixed': True,
            'unit': val.to_dict()['unit']
        } for lbl, val in self.fixed.items()})
        result_dict = self.eval_constrained_results(result_dict, self.constrained)

        r_squared_args = self.x_data_reduced, self.y_data, self.x_data_reducer, self.observer.system_cls
        r_dict = {key: value['value'] for key, value in result_dict.items()}

        r_squared_result = rv_r_squared(rv_model.central_rv_synthetic, *r_squared_args, **r_dict)
        result_dict["r_squared"] = {'value': r_squared_result, "unit": None}

        setattr(self, 'flat_result', result_dict)
        return parameters.serialize_result(result_dict)



from . dataset.base import RVData, LCData
from .. base.transform import TransformProperties
from .. import settings


class RVBinaryAnalyticsTask(TransformProperties):
    """
    Evaluating whether the observational data in elisa.tasks.RVBinaryAnalyticsTask are supplied in valid form.
    """
    @staticmethod
    def data(value):
        if isinstance(value, dict):
            for key, val in value.items():
                if key not in settings.BINARY_COUNTERPARTS:
                    raise ValueError(f'{key} is invalid designation for radial velocity dataset. '
                                     f'Please choose from {settings.BINARY_COUNTERPARTS.keys()}')
                elif not isinstance(val, RVData):
                    raise TypeError(f'{val} is not of instance of RVData class.')
            return value
        raise TypeError('`radial_velocities` are not of type `dict`')


class LCBinaryAnalyticsProperties(TransformProperties):
    """
    Evaluating whether the observational data in elisa.tasks.LCBinaryAnalyticsTask are supplied in valid form.
    """
    @staticmethod
    def data(value):
        if isinstance(value, dict):
            for key, val in value.items():
                if key not in settings.PASSBANDS:
                    raise ValueError(f'{key} is invalid passband. Please choose '
                                     f'from available passbands: \n{settings.PASSBANDS}')
                elif not isinstance(val, LCData):
                    raise TypeError(f'{val} is not of instance of LCData class.')
            return value
        raise TypeError('`light_curves` are not of type `dict`')



from ... base.curves import utils as crv_utils
from ... import settings


def _calculate_lc_point(band, system):
    """
    Calculates point on the light curve for given band.

    :param band: str; name of the photometric band compatibile with supported names in config
    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :return: float;
    """
    flux = 0.0
    for component in settings.BINARY_COUNTERPARTS:
        star = getattr(system, component)
        flux += crv_utils.flux_from_star_container(band, star)
    return flux


def compute_lc_on_pos(band_curves, pos_idx, passbands, system):
    """
    Calculates lc points for given orbital position.

    :param band_curves: Dict; {str; passband : numpy.array; light curve, ...} result will be written to the
                              corresponding `pos_idx` position
    :param pos_idx: int; position in `band_curves` to which calculated lc points will be assigned
    :param passbands: List; list of passbands
    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :return: Dict; updated {str; passband : numpy.array; light curve, ...}
    """
    # integrating resulting flux
    for band in passbands:
        band_curves[band][pos_idx] = _calculate_lc_point(band, system)
    return band_curves



import numpy as np

from .. import utils as butils, dynamic
from ... base import transform
from ... base.graphics import plot
from .. container import OrbitalPositionContainer
from ... const import Position
from ... graphic import graphics
from ... base.surface.faces import correct_face_orientation
from .. curves import utils as crv_utils
from ... observer.observer import Observer

from ... import (
    umpy as up,
    units as u,
    utils,
    const
)


class Plot(object):
    """
    Universal plot interface for binary system class, more detailed documentation for each value of descriptor is
    available in graphics library.
    
    available plot methods::
                        
        `orbit` - plots orbit in orbital plane
        `equipotential` - plots crossections of surface Hill planes in xy,yz,zx planes
        `mesh` - plot surface points
        `wireframe` - wire frame model of the binary
        `surface` - plot stellar surfaces
    """

    defpos = Position(*(0, 1.0, 0.0, 0.0, 0.0))

    def __init__(self, instance):
        self.binary = instance

    def orbit(self, start_phase=0.0, stop_phase=1.0, number_of_points=300,
              axis_units=u.solRad, frame_of_reference='primary', legend=True, return_figure_instance=False):
        """
        Function for quick 2D plot of the orbital motion in the orbital plane.

        :param start_phase: float; starting phase for the plot
        :param stop_phase: float; finishing phase for the plot
        :param number_of_points: int; number of points in the plot
        :param axis_units: Union[astropy.unit, str]; specifying axis unit, use astropy
                           units or `dimensionless` or `SMA` (semi-major axis) units for axis scale
        :param frame_of_reference: str; `barycentric` or `primary`
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        :param legend: bool;
        """
        orbit_kwargs = dict()
        if axis_units in ['dimensionless', 'SMA']:
            axis_units = u.dimensionless_unscaled

        # orbit calculation for given phases
        phases = np.linspace(start_phase, stop_phase, number_of_points)
        ellipse = self.binary.orbit.orbital_motion(phase=phases)

        # if axis are without unit a = 1
        if axis_units != u.dimensionless_unscaled:
            a = self.binary.semi_major_axis * u.DefaultBinarySystemUnits.system.semi_major_axis.to(axis_units)
            radius = a * ellipse[:, 0]
        else:
            radius = ellipse[:, 0]
        azimuth = ellipse[:, 1]
        x, y = utils.polar_to_cartesian(radius=radius, phi=azimuth + const.PI / 2.0)

        if frame_of_reference == 'barycentric':
            orbit_kwargs.update({
                'x1_data': - self.binary.mass_ratio * x / (1 + self.binary.mass_ratio),
                'y1_data': - self.binary.mass_ratio * y / (1 + self.binary.mass_ratio),
                'x2_data': x / (1 + self.binary.mass_ratio),
                'y2_data': y / (1 + self.binary.mass_ratio)
            })
        elif frame_of_reference == 'primary':
            orbit_kwargs.update({
                'x_data': x,
                'y_data': y
            })
        orbit_kwargs.update({
            "return_figure_instance": return_figure_instance,
            "axis_units": axis_units,
            "start_phase": start_phase,
            "stop_phase": stop_phase,
            "number_of_points": number_of_points,
            "frame_of_reference": frame_of_reference,
            "legend": legend
        })
        return graphics.orbit(**orbit_kwargs)

    def equipotential(self, plane='xy', phase=0.0, components_to_plot='both', colors=('b', 'r'), legend=True,
                      legend_loc=1, return_figure_instance=False):
        """
        Function for quick 2D plot of equipotential cross-section.

        :param plane: str; (`xy`, `yz` or `xz`) specifying what plane cross-section to display default is `xy`
        :param phase: float; phase at which to plot cross-section
        :param components_to_plot: str; component to plot `primary`, `secondary` or `both` (default)
        :param colors: Tuple; tuple of colors for primary and secondary component equipotentials
        :param legend: bool; legend display on/off
        :param legend_loc: int; location of the legend
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        """
        equipotential_kwargs = dict()
        # relative distance between components (a = 1)
        if utils.is_plane(plane, 'xy') or utils.is_plane(plane, 'yz') or utils.is_plane(plane, 'zx'):
            components_distance = self.binary.orbit.orbital_motion(phase=phase)[0][0]
            points_primary, points_secondary = self.binary.compute_equipotential_boundary(components_distance, plane)
        else:
            raise ValueError('Invalid choice of crossection plane, use only: `xy`, `yz`, `zx`.')

        equipotential_kwargs.update({
            "return_figure_instance": return_figure_instance,
            "plane": plane,
            "phase": phase,
            "points_primary": points_primary,
            "points_secondary": points_secondary,
            "components_to_plot": components_to_plot,
            "colors": colors,
            "legend": legend,
            "legend_loc": legend_loc
        })
        return graphics.equipotential(**equipotential_kwargs)

    def mesh(self, phase=0.0, components_to_plot='both', plot_axis=True, inclination=None, azimuth=None,
             return_figure_instance=False):
        """
        Function plots 3D scatter plot of the surface points.

        :param plot_axis: bool; switch the plot axis on/off
        :param inclination: Union[float, astropy.Quantity]; elevation of the camera (in degrees if float)
        :param azimuth: Union[float, astropy.Quantity]; azimuth of the camera (in degrees if float)
        :param components_to_plot: str; component to plot `primary`, `secondary` or `both` (default)
        :param phase: float; phase at which to construct plot
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        """

        binary_mesh_kwargs = dict()
        inclination = transform.deg_transform(inclination, u.deg, transform.WHEN_FLOAT64,
                                              u.DefaultBinarySystemInputUnits.system.inclination) \
            if inclination is not None else up.degrees(self.binary.inclination)
        components_distance, azim = self.binary.orbit.orbital_motion(phase=phase)[0][:2]

        azimuth = transform.deg_transform(azimuth, u.deg, when_float64=transform.WHEN_FLOAT64) \
            if azimuth is not None else up.degrees(azim) - 90

        orbital_position_container = OrbitalPositionContainer.from_binary_system(self.binary, self.defpos)
        orbital_position_container.build_mesh(components_distance=components_distance)
        orbital_position_container.build_perturbations(component=components_to_plot,
                                                       components_distance=components_distance)

        if components_to_plot in ['primary', 'both']:
            binary_mesh_kwargs.update({
                'points_primary': orbital_position_container.primary.get_flatten_parameter('points')
            })

        if components_to_plot in ['secondary', 'both']:
            binary_mesh_kwargs.update({
                'points_secondary': orbital_position_container.secondary.get_flatten_parameter('points')
            })

        binary_mesh_kwargs.update({
            "return_figure_instance": return_figure_instance,
            "phase": phase,
            "components_to_plot": components_to_plot,
            "plot_axis": plot_axis,
            "inclination": inclination,
            "azimuth": azimuth,
        })
        return graphics.binary_mesh(**binary_mesh_kwargs)

    def wireframe(self, phase=0.0, components_to_plot='both', plot_axis=True, inclination=None, azimuth=None,
                  return_figure_instance=False):
        """
        Function displays wireframe model of the stellar surface.

        :param phase: float; phase at which to construct plot
        :param components_to_plot: str; component to plot `primary`, `secondary` or `both` (default)
        :param plot_axis: bool; switch the plot axis on/off
        :param inclination: Union[float, astropy.Quantity]; elevation of the camera (in degrees if float)
        :param azimuth: Union[float, astropy.Quantity]; azimuth of the camera (in degrees if float)
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        """
        binary_wireframe_kwargs = dict()
        inclination = transform.deg_transform(inclination, u.deg, transform.WHEN_FLOAT64,
                                              u.DefaultBinarySystemInputUnits.system.inclination) \
            if inclination is not None else up.degrees(self.binary.inclination)
        components_distance, azim = self.binary.orbit.orbital_motion(phase=phase)[0][:2]
        azimuth = transform.deg_transform(azimuth, u.deg, when_float64=transform.WHEN_FLOAT64) \
            if azimuth is not None else up.degrees(azim) - 90

        orbital_position_container = OrbitalPositionContainer.from_binary_system(self.binary, self.defpos)

        # recalculating spot latitudes
        spots_longitudes = dynamic.calculate_spot_longitudes(self.binary, phase, component="all")
        dynamic.assign_spot_longitudes(orbital_position_container, spots_longitudes, index=None, component="all")

        orbital_position_container.build_mesh(components_distance=components_distance)
        orbital_position_container.build_faces(components_distance=components_distance)

        if components_to_plot in ['primary', 'both']:
            points, faces = orbital_position_container.primary.surface_serializer()
            binary_wireframe_kwargs.update({
                "points_primary": points,
                "primary_triangles": faces
            })
        if components_to_plot in ['secondary', 'both']:
            points, faces = orbital_position_container.secondary.surface_serializer()
            binary_wireframe_kwargs.update({
                "points_secondary": points,
                "secondary_triangles": faces
            })
        binary_wireframe_kwargs.update({
            "return_figure_instance": return_figure_instance,
            "phase": phase,
            "components_to_plot": components_to_plot,
            "plot_axis": plot_axis,
            "inclination": inclination,
            "azimuth": azimuth
        })
        return graphics.binary_wireframe(**binary_wireframe_kwargs)

    def surface(self, phase=0.0, components_to_plot='both', normals=False, edges=False, colormap=None, plot_axis=True,
                face_mask_primary=None, face_mask_secondary=None, elevation=None, azimuth=None, colorbar_unit='default',
                axis_unit=u.dimensionless_unscaled, colorbar_orientation='vertical', colorbar=True, scale='linear',
                surface_colors=('g', 'r'), separate_colormaps=None, colorbar_separation=0.0, colorbar_size=0.7,
                return_figure_instance: bool=False, subtract_equilibrium: bool=False):
        """
        Function creates plot of binary system components

        :param phase: float; phase at which plot the system
        :param components_to_plot: str; `primary`, `secondary` or `both` (default),
        :param normals: bool; plot normals of the surface phases as arrows
        :param edges: bool; highlight edges of surface faces
        :param colormap: str;
        :param plot_axis: bool; if False, axis will be hidden
        :param face_mask_primary: array[bool]; mask to select which faces to display
        :param face_mask_secondary: array[bool]: mask to select which faces to display
        :param elevation: Union[float, astropy.Quantity]; in degrees - elevation of camera
        :param azimuth: Union[float, astropy.Quantity]; camera azimuth
        :param colorbar_unit: str; colorbar unit
        :param axis_unit: Union[astropy.unit, dimensionless]; - axis units
        :param colorbar_orientation: str; 'horizontal' or 'vertical' (default)
        :param colorbar: bool; colorbar on/off switch
        :param scale: str; 'linear' or 'log'
        :param surface_colors: Tuple; tuple of colors for components if `colormap` are not specified
        :param separate_colormaps: bool; if True, figure will contain separate colormap for each component
        :param colorbar_separation: float; shifting position of the colorbar from its default postition, default is 0.0
        :param colorbar_size: float; relative size of the colorbar, default 0.7
        :param subtract_equilibrium: bool; if True; equilibrium values are subtracted from surface colormap
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        :**available colormap options**:
            * :'gravity_acceleration': surface distribution of gravity acceleration,
            * :'temperature': surface distribution of the effective temperature,
            * :'velocity': absolute values of surface elements velocities with respect to the observer,
            * :'radial_velocity': radial component of the surface element velocities relative to the observer,
            * :'normal_radiance': surface element radiance perpendicular to the surface element,
            * :'radiance': radiance of the surface element in a direction towards the observer,
            * :'radius': distance of the surface elements from the centre of mass
            * :'horizontal_displacement': distribution of the horizontal component of surface displacement
            * :'horizontal_acceleration': distribution of horizontal component surface acceleration
            * :'v_r_perturbed': radial component of the pulsation velocity (perpendicular towards
            * :'v_horizontal_perturbed': horizontal component of the pulsation  velocity

        :return: Union[None, matplotlib.pyplot.figure]
        """
        surface_kwargs = dict()

        elevation = transform.deg_transform(elevation, u.deg, when_float64=transform.WHEN_FLOAT64) \
            if elevation is not None else 0
        
        orbital_position = self.binary.orbit.orbital_motion(phase=phase)[0]
        components_distance, azim = orbital_position[:2]
        orbital_position = Position(0, orbital_position[0], orbital_position[1],
                                    orbital_position[2], orbital_position[3])
        
        azimuth = transform.deg_transform(azimuth, u.deg, when_float64=transform.WHEN_FLOAT64) \
            if azimuth is not None else up.degrees(azim) - 90

        if separate_colormaps is None:
            separate_colormaps = self.binary.morphology != 'over-contact' and \
                                 colormap not in ['velocity', 'radial_velocity'] and components_to_plot == 'both'

        potentials = self.binary.correct_potentials([phase, ], component="all", iterations=2)
        
        orbital_position_container = OrbitalPositionContainer.from_binary_system(self.binary, self.defpos)
        # recalculating spot latitudes
        spots_longitudes = dynamic.calculate_spot_longitudes(self.binary, phase, component="all")
        dynamic.assign_spot_longitudes(orbital_position_container, spots_longitudes, index=None, component="all")
        
        orbital_position_container.set_on_position_params(orbital_position, potentials["primary"][0],
                                                          potentials["secondary"][0])

        orbital_position_container.set_time()
        orbital_position_container.build(components_distance=components_distance, components='both',
                                         build_pulsations=True)
        components = butils.component_to_list(components_to_plot)

        com = {'primary': 0.0, 'secondary': components_distance}
        for component in components:
            correct_face_orientation(getattr(orbital_position_container, component), com=com[component])

        # calculating radiances
        o = Observer(passband=['bolometric', ], system=self.binary)
        atm_kwargs = dict(
            passband=o.passband,
            left_bandwidth=o.left_bandwidth,
            right_bandwidth=o.right_bandwidth,
        )

        crv_utils.prep_surface_params(
            system=orbital_position_container, write_to_containers=True, **atm_kwargs
        )

        orbital_position_container = butils.move_sys_onpos(orbital_position_container, orbital_position, on_copy=True)
        args = (orbital_position.distance, self.binary.mass_ratio, orbital_position_container.secondary.com)
        pos_correction = butils.correction_to_com(*args)

        for component in components:
            star = getattr(orbital_position_container, component)

            args = (colormap, star, phase, com[component], self.binary.semi_major_axis, self.binary.inclination,
                    orbital_position_container.position)
            kwargs = dict(scale=scale, unit=colorbar_unit, subtract_equilibrium=subtract_equilibrium)

            surface_kwargs.update({f'{component}_cmap': plot.add_colormap_to_plt_kwargs(*args, **kwargs)})

            surface_kwargs.update({
                f'points_{component}': star.points - pos_correction[None, :],
                f'{component}_triangles': star.faces
            })

            face_mask = locals().get(f'face_mask_{component}')
            face_mask = np.ones(star.faces.shape[0], dtype=bool) if face_mask is None else face_mask
            surface_kwargs[f'{component}_triangles'] = surface_kwargs[f'{component}_triangles'][face_mask]
            if colormap is not None:
                surface_kwargs[f'{component}_cmap'] = surface_kwargs[f'{component}_cmap'][face_mask]

            if normals:
                surface_kwargs.update({
                    f'{component}_centres': star.face_centres[face_mask] - pos_correction[None, :],
                    f'{component}_arrows': star.normals[face_mask]
                })

            if axis_unit != u.dimensionless_unscaled:
                sma = (self.binary.semi_major_axis * u.DefaultBinarySystemUnits.system.semi_major_axis).\
                    to(axis_unit).value
                surface_kwargs[f'points_{component}'] *= sma

                if normals:
                    surface_kwargs[f'{component}_centres'] *= sma
                    surface_kwargs[f'{component}_arrows'] *= sma

        surface_kwargs.update({
            "phase": phase,
            "components_to_plot": components_to_plot,
            "normals": normals,
            "edges": edges,
            "colormap": colormap,
            "plot_axis": plot_axis,
            "face_mask_primary": face_mask_primary,
            "face_mask_secondary": face_mask_secondary,
            "elevation": elevation,
            "azimuth": azimuth,
            "unit": colorbar_unit,
            "axis_unit": axis_unit,
            "colorbar_orientation": colorbar_orientation,
            "colorbar": colorbar,
            "scale": scale,
            "morphology": self.binary.morphology,
            "surface_colors": surface_colors,
            "separate_colormaps": separate_colormaps,
            'colorbar_separation': colorbar_separation,
            'colorbar_size': colorbar_size,
            'return_figure_instance': return_figure_instance
        })

        return graphics.binary_surface(**surface_kwargs)



__version__ = '0.6.0'

import json
import os.path as op
import sys

from . conf.settings import settings
from . binary_system.system import BinarySystem
from . managers.download_manager import DownloadManager
from . single_system.system import SingleSystem
from . base.star import Star
from . observer.observer import Observer
from . analytics.dataset.base import LCData, RVData
from . import units


def get_default_binary_definition():
    _path = op.join(op.dirname(__file__), "data", "default_binary_system.json")
    with open(_path, "r") as f:
        return json.loads(f.read())


def _default_binary():
    return BinarySystem.from_json(data=get_default_binary_definition())


def _default_observer():
    return Observer(passband=["Generic.Bessell.U", "Generic.Bessell.V", "Generic.Bessell.R"], system=_default_binary())


def _bolometric_default_observer():
    return Observer(passband=["bolometric"], system=_default_binary())


get_default_binary = _default_binary
get_default_observer = _default_observer
get_bolometric_default_observer = _bolometric_default_observer

# prepare units as u for simpler import
u = units

# donwload manager
download_manager = DownloadManager(settings)

# first time user
if settings.FIRST_TIME_USER:

    download = input("Download manager will pull atmospheres and limb darkening tables.\n"
                     "Do you want to proceed? [y/N]: ")
    if not download.lower() == "y":
        print("Please use download manager to pull atmospheres \n "
              "and limb darkening or do it manually as refered docs.")
        sys.exit(0)
    download_manager.download_all()



import numpy as np

from .. import umpy as up
from .. base import error


def newton(func, x0, fprime, args=(), maxiter=50, rtol=0.0):
    """
    Newton root solver.
    Solved by approach::

        x_n+1 = x_n - (func(x_n) / fprime(x_n))

    Starts from x0.

    :param func: callable; the function whose zero is wanted. It must be a function of a single variable
                           of the form f(x,a,b,c...), where a, b, c... are extra arguments that can be passed
                           in the args parameter.
    :param x0: Union[numpy.array, float]; an initial estimate of the solution
    :param fprime: callable; the derivative of the function
    :param args: Tuple; extra arguments to be used in the function call
    :param maxiter: int; maximum number of iterations
    :param rtol: tolerance (relative) for termination
    :return: Union[numpy.array, float]; depends on the `x0` shape
    """
    x_n = np.copy(x0) if isinstance(x0, np.ndarray) else x0
    for _ in range(maxiter):
        difference = func(x_n, *args) / fprime(x_n, *args[0])
        x_m = x_n - difference
        if np.max(up.abs(difference / x_n)) <= rtol:
            return x_m
        x_n = x_m
    raise error.MaxIterationError(f"Max iteration limit - {maxiter} - exceeded")
